from __future__ import annotations

import typing

import toolstr

from ctc import spec
from ctc import cli
from . import yearn_spec


def get_yearn_addresses(
    network: spec.NetworkReference,
) -> typing.Mapping[str, spec.Address]:

    if network not in ('ethereum', 1):
        raise NotImplementedError('non-ethereum address')

    mainnnet_addresses = {
        'Oracle': '0x83d95e0d5f402511db06817aff3f9ea88224b030',
        'ManagementList': '0xf64e58ee8c7badc741a7ea98fb65488084385674',
        #
        # registries
        'AdapterV2': '0x240315db938d44bb124ae619f5fd0269a02d1271',
        'AdapterIronBank': '0xff0bd2d0c7e9424ccb149ed3757155eef41a793d',
        #
        # tvls
        'TvlV2': '0x14d6e0908bae40a2487352b2a9cb1a6232da8785',
        'TvlV1': '0xf4fb8903a41fc78686b26de55502cde42a4c6c78',
        'TvlIronBank': '0xec7ac8ac897f5082b2c3d4e8d2173f992a097f24',
        'TvlEarn': '0x1007ed6fdfac72bbea9c719cf1fa9c355d248691',
        'TvlveCRV': '0x560144c25e53149ac410e5d33bdb131e49a850e5',
        #
        # helpers
        'Helper': '0x5aacd0d03096039ac4381cd814637e9fb7c34a6f',
        'UniqueAddressesHelper': '0xc3a0bef4a47ba579cbba510ae2c59d9b9bf9467c',
        'StrategiesHelper': '0xae813841436fe29b95a14ac701afb1502c4cb789',
        'AllowancesHelper': '0x4218e20db87023049fc582aaa4bd47a3611a20ab',
        'AddressMergeHelper': '0x957e3ae7983155a9f9e08da555b8084448be26e4',
        'PricesHelper': '0x5d63a8584d91ebc5033d022afd6c5a7c7fddc99b',
        'BalancesHelper': '0x855ffe28019106d089bc018df18838f8d241c402',
        #
        # generators
        'AddressesGeneratorV1Vaults': '0xce29d34c8e88a2e1edde10ad4eee4f3e379fc041',
        'AddressesGeneratorV2Vaults': '0x437758d475f70249e03eda6be23684ad1fc375f0',
        'AddressesGeneratorEarn': '0xf536399c04683d7ab0bcb1308c245b4bbb539344',
        'AddressesGeneratorIronBank': '0xa0b57619a980dfefd50f24f310ee1b55a40a9d46',
        #
        # delegation mapping
        'DelegatedBalanceMapping': '0xc01a529c01f9399c928a4afc50e25f12d1e5b142',
        #
        # calculations
        'CalculationsCurve': '0xe9cff16bdac9729f18cbac451ecca6c50b372207',
        'CalculationsIronBank': '0x55e9b18fefff7e00548d54480373fc8843de8ea4',
        'CalculationsSushiswap': '0x88de7d7f7b9597c86b8cd195374fbf602934f334',
        'CalculationsYearnVaults': '0x38477f2159638956d33e18951d98238a53b9aa3c',
        #
        # registries
        'RegistryEarn': '0x62a4e0e7574e5407656a65cc8dbdf70f3c6eb04b',
        'RegistryV2': '0x50c1a2ea0a861a967d9d0ffe2ae4012c2e053804',
        'RegistryV1': '0x3ee41c098f9666ed2ea246f4d2558010e59d63a0',
        'RegistryIB': '0xab1c342c7bf5ec5f02adea1c2270670bca144cbb',
    }

    return mainnnet_addresses


def get_yearn_address(
    name: str,
    network: spec.NetworkReference,
) -> spec.Address:

    addresses = get_yearn_addresses(network=network)
    return addresses[name]


#
# # address printing
#


def print_lens_addresses(network: spec.NetworkReference) -> None:
    addresses = get_yearn_addresses(network=network)
    rows = []
    for name, address in addresses.items():
        row = [name, address]
        rows.append(row)
    labels = [
        'name',
        'address',
    ]
    styles = cli.get_cli_styles()
    toolstr.print_table(
        rows,
        labels=labels,
        border=styles['comment'],
        label_style=styles['title'],
        column_styles={
            'name': styles['option'],
            'address': styles['metavar'],
        },
        indent=4,
    )


def print_vault_addresses(
    api_vaults: typing.Sequence[yearn_spec.ApiVault], verbose: bool = False
) -> None:
    if verbose:
        _print_vault_addresses_verbose(api_vaults=api_vaults)
    else:
        _print_vault_addresses_nonverbose(api_vaults=api_vaults)


def _print_vault_addresses_nonverbose(
    api_vaults: typing.Sequence[yearn_spec.ApiVault],
) -> None:
    rows = []
    for datum in api_vaults:
        row = [
            datum['name'],
            datum['address'].lower(),
        ]
        rows.append(row)

    labels = [
        'vault',
        'address',
    ]
    sort_index = labels.index('vault')
    rows = sorted(rows, key=lambda row: row[sort_index])

    styles = cli.get_cli_styles()
    toolstr.print_table(
        rows,
        labels=labels,
        border=styles['comment'],
        label_style=styles['title'],
        column_styles={
            'vault': styles['option'],
            'address': styles['metavar'],
        },
        indent=4,
    )


def _print_vault_addresses_verbose(
    api_vaults: typing.Sequence[yearn_spec.ApiVault],
) -> None:

    styles = cli.get_cli_styles()

    rows = []
    for datum in api_vaults:

        addresses = {
            'vault': datum['address'],
            'underlying': datum['token']['address'],
        }
        for s, strategy in enumerate(datum['strategies']):
            addresses['strategy ' + str(s)] = strategy['address']
        address_pieces = [
            toolstr.add_style(name, styles['comment'])
            + ' '
            + toolstr.add_style(address.lower(), styles['metavar'])
            for name, address in addresses.items()
        ]

        row = [datum['name'], '\n'.join(address_pieces)]
        rows.append(row)

    labels = [
        'vault',
        'address',
    ]
    sort_index = labels.index('vault')
    rows = sorted(rows, key=lambda row: row[sort_index])

    toolstr.print_multiline_table(
        rows,
        labels=labels,
        border=styles['comment'],
        label_style=styles['title'],
        column_styles={
            'vault': styles['description'],
            'address': styles['metavar'],
        },
    )



from __future__ import annotations


registry = '0x00000000000c2e074ec69a0dfb2997ba6c7d2e1e'
base_registrar = (
    '0x57f1887a8bf19b14fc0df6fd9b2acc9af147ea85'  # the erc721 token
)
resolver = '0x4976fb03c32e5b8cfe2b6ccb31c09ba78ebaba41'
reverse_records = '0x3671ae578e63fdf66ad4f3e12cc0c0d71ac7510c'
controller = '0x283af0b28c62c092c9727f1ee09c02ca627eb7f5'


old_resolvers = [
    '0x226159d592E2b063810a10Ebf6dcbADA94Ed68b8',
    '0xdaaf96c344f63131acadd0ea35170e7892d3dfba',
]



from __future__ import annotations

import typing

import toolsql

from ctc import spec
from . import coingecko_schema_defs


async def async_upsert_tokens(
    *,
    tokens: typing.Sequence[coingecko_schema_defs.CoingeckoToken],
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> None:

    if len(tokens) == 0:
        return

    await toolsql.async_insert(
        conn=conn,
        table='coingecko_tokens',
        rows=tokens,
        upsert=True,
    )


async def async_delete_tokens(
    *,
    ids: typing.Sequence[str],
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> None:

    if len(ids) == 0:
        return

    await toolsql.async_delete(
        conn=conn,
        table='coingecko_tokens',
        where_in={'id': ids},
    )


async def async_select_token(
    *,
    conn: toolsql.AsyncConnection,
    id: str | None = None,
    context: spec.Context = None,
) -> coingecko_schema_defs.CoingeckoToken | None:

    if id is not None:
        where_equals = {'id': id}
    else:
        raise Exception('must specify id')

    result: coingecko_schema_defs.CoingeckoToken = await toolsql.async_select(  # type: ignore
        conn=conn,
        table='coingecko_tokens',
        where_equals=where_equals,
        output_format='single_dict',
    )

    return result


async def async_select_tokens(
    *,
    symbol_query: str | None = None,
    name_query: str | None = None,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> typing.Sequence[coingecko_schema_defs.CoingeckoToken] | None:

    # symbol query
    if symbol_query is not None:
        symbol_filter: toolsql.WhereGroup | None = {
            'where_ilike': {'symbol': '%' + symbol_query.lower() + '%'}
        }
    else:
        symbol_filter = None

    # name query
    if name_query is not None:
        name_filter: toolsql.WhereGroup | None = {
            'where_ilike': {'symbol': '%' + name_query.lower() + '%'}
        }
    else:
        name_filter = None

    # combine filters
    if symbol_filter is not None and name_filter is not None:
        select_kwargs: toolsql.SelectKwargs = {'where_or': [symbol_filter, name_filter]}
    elif symbol_filter is not None:
        select_kwargs = symbol_filter  # type: ignore
    elif name_filter is not None:
        select_kwargs = name_filter  # type: ignore
    else:
        select_kwargs = {}

    result: typing.Sequence[
        coingecko_schema_defs.CoingeckoToken
    ] = await toolsql.async_select(  # type: ignore
        conn=conn,
        table='coingecko_tokens',
        order_by='market_cap_rank',
        **select_kwargs,
    )

    ranked: list[coingecko_schema_defs.CoingeckoToken] = []
    unranked: list[coingecko_schema_defs.CoingeckoToken] = []
    for item in result:
        if item['market_cap_rank'] is not None:
            ranked.append(item)
        else:
            unranked.append(item)

    together: typing.MutableSequence[coingecko_schema_defs.CoingeckoToken] = (
        ranked + unranked
    )

    return together




import pytest

from ctc import evm


@pytest.mark.asyncio
async def test_get_block():
    latest_block = await evm.async_get_block('latest')
    assert latest_block is not None


@pytest.mark.asyncio
async def test_fetch_latest_block_number():
    latest_block_number = await evm.async_get_latest_block_number()
    assert isinstance(latest_block_number, int)


@pytest.mark.asyncio
async def test_get_contract_creation_block():
    contract = '0x6B175474E89094C44Da98b954EedeAC495271d0F'
    result = await evm.async_get_contract_creation_block(contract)
    target = 8928158
    assert result == target


block_timestamps = {
    4019263: 1500000000,
    10853627: 1600000000,
}


@pytest.mark.asyncio
@pytest.mark.parametrize('block_timestamp', block_timestamps.items())
async def test_get_block_of_timestamp(block_timestamp):
    block, timestamp = block_timestamp
    obtained_block = await evm.async_get_block_of_timestamp(
        timestamp,
        context={'cache': True},
    )
    assert block == obtained_block

    obtained_block = await evm.async_get_block_of_timestamp(
        timestamp,
        context={'cache': False},
    )
    assert block == obtained_block


@pytest.mark.asyncio
@pytest.mark.parametrize('block_timestamp', block_timestamps.items())
@pytest.mark.parametrize(
    'db_settings', [[False, False], [False, True], [True, True]]
)
async def test_block_of_timestamp_modes(block_timestamp, db_settings):
    block, _ = block_timestamp
    use_db, use_db_assist = db_settings
    context = {'cache': use_db}

    block_data = await evm.async_get_block(block)

    timestamp = block_data['timestamp']
    timestamp_after = timestamp + 1

    # test using exact timestamp

    obtained_block = await evm.async_get_block_of_timestamp(
        timestamp, mode='==', context=context
    )
    assert obtained_block == block

    obtained_block = await evm.async_get_block_of_timestamp(
        timestamp, mode='<=', context=context
    )
    assert obtained_block == block

    obtained_block = await evm.async_get_block_of_timestamp(
        timestamp, mode='>=', context=context
    )
    assert obtained_block == block

    # test using offset timestamp
    with pytest.raises(Exception):
        obtained_block = await evm.async_get_block_of_timestamp(
            timestamp_after,
            mode='==',
            context=context
        )

    obtained_block = await evm.async_get_block_of_timestamp(
        timestamp_after, mode='<=', context=context
    )
    assert obtained_block == block

    obtained_block = await evm.async_get_block_of_timestamp(
        timestamp_after, mode='>=', context=context
    )
    assert obtained_block == block + 1



"""
# TODO
- amount burned/issued
    - need to sum up
        - minter rewards (both pre and post 1559)
        - uncle inclusion rewards
        - uncle mining rewards
    - see https://github.com/ethereum/go-ethereum/issues/1644
- amount of time in each block aggregation range
- historical blocks and block ranges
- color


use statistical sampling to get quick daily/weekly/monthly avg's
"""

from __future__ import annotations

import time
import typing

import toolcli
import toolstr
import tooltime

import ctc.config
from ctc import cli
from ctc import evm
from ctc import rpc
from ctc.cli import cli_utils


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_gas_command,
        'help': 'output gas summary of block range',
        'args': [
            # {
            #     'name': 'last',
            #     'help': 'either amount of blocks (e.g. 100) or time (e.g. 24h)',
            #     'nargs': '?',
            # },
            {
                'name': '--block',
                'help': 'historical block to use',
                'type': int,
            },
            {
                'name': '--last',
                'metavar': 'N',
                'nargs': '+',
                'help': 'number of blocks to include',
            },
            {
                'name': '--export',
                'default': 'stdout',
                'help': 'file path for output (.json or .csv)',
            },
            {
                'name': '--overwrite',
                'action': 'store_true',
                'help': 'specify that output path can be overwritten',
            },
            {
                'name': ['-v', '--verbose'],
                'action': 'store_true',
                'help': 'output additional information',
            },
        ],
        'examples': [
            '',
            '--last 30',
        ],
    }


async def async_gas_command(
    *,
    block: int | None,
    last: typing.Optional[typing.Sequence[str]],
    export: str,
    overwrite: bool,
    verbose: bool,
) -> None:

    import numpy as np

    styles = cli.get_cli_styles()

    if last is None:
        last_as_int = [10, 100]
    else:
        last_as_int = [
            int(subtoken.strip(' '))
            for token in last
            for subtoken in token.strip(',').split(',')
        ]

    n_blocks = max(last_as_int)
    if block is None:
        block = await rpc.async_eth_block_number()
        using_latest = True
    else:
        using_latest = False
    block_numbers = list(range(block - n_blocks + 1, block + 1))

    # get block transaction data
    raw_median_gas_fees = await evm.async_get_median_blocks_gas_fees(
        block_numbers,
        latest_block_number=block,
    )

    median_gas_fees = np.array(
        [item['median_gas_fee'] for item in raw_median_gas_fees],
        dtype=float,
    )
    # block_timestamps = await evm.async_get_block_timestamps(block_numbers)
    block_timestamps = [item['timestamp'] for item in raw_median_gas_fees]

    if using_latest:
        now = time.time()
    else:
        now = block_timestamps[-1]
    last_times = ['5 minutes', '10 minutes']
    last_blocks = []
    for last_as_int_time in last_times:
        cutoff_timestamp = now - tooltime.timelength_to_seconds(
            last_as_int_time
        )
        for i in range(n_blocks):
            if block_timestamps[-i - 1] < cutoff_timestamp:
                break
        last_blocks.append(i)

    toolstr.print_text_box('Gas Price Summary', style=styles['title'])
    print()
    toolstr.print('all prices reported in gwei', style=styles['comment'])

    print()
    toolstr.print_header('Latest Block Summary', style=styles['title'])
    print()
    toolstr.print(
        'using block ' + toolstr.add_style(str(block), styles['metavar']),
        style=styles['comment'],
    )
    block_rows = []
    block_gas_stats = await evm.async_get_block_gas_stats(
        block_numbers[-1],
    )
    for key, value in block_gas_stats.items():
        if key in ['gas_limit', 'gas_used']:
            block_row = (
                key,
                toolstr.format(value, order_of_magnitude=True),
            )
        else:
            if value is None:
                if key == 'base_fee':
                    block_row = (key, '-')
                else:
                    block_row = (key, '\[no txs]')
            else:
                block_row = (key, toolstr.format(value))
        block_rows.append(block_row)
    print()
    toolstr.print_table(
        block_rows,
        border=styles['comment'],
        column_styles=[styles['option'], styles['description'] + ' bold'],
        indent=4,
    )

    if verbose:
        print()
        print()
        toolstr.print_header(
            'Costs of Common Operations', style=styles['title']
        )
        from ctc.protocols import chainlink_utils

        recent_median_gas_fee = median_gas_fees[~np.isnan(median_gas_fees)][-1]

        try:
            eth_usd: float | None = await chainlink_utils.async_get_eth_price(
                block=block
            )
        except Exception:
            eth_usd = None
        print()
        message = 'using gas price = ' + toolstr.add_style(
            toolstr.format(recent_median_gas_fee) + ' gwei',
            styles['description'] + ' bold',
        )
        if eth_usd is not None:
            message += ' and ETH price = ' + toolstr.add_style(
                toolstr.format(eth_usd, prefix='$'),
                styles['description'] + ' bold',
            )
        toolstr.print(message, style=styles['comment'])
        print()
        await async_print_gas_costs(
            eth_usd=eth_usd,
            gas_fee=recent_median_gas_fee,
        )

    print()
    print()
    toolstr.print_header('Previous Blocks', style=styles['title'])
    print()
    toolstr.print(
        'aggregated by median transaction price of each block',
        style=styles['comment'],
    )
    print()
    labels = ['', 'blocks', 'time', 'min', 'median', 'mean', 'max']
    rows: typing.MutableSequence[typing.Any] = []
    for index, last_n in enumerate(last_as_int + last_blocks):

        row: list[typing.Union[str, None, int, float]] = []

        if index < len(last_as_int):
            if last_n == 1:
                row.append('latest block')
            else:
                row.append('last ' + str(last_n) + ' blocks')
            row.append(last_n)
            if using_latest:
                timelength_seconds = round(now - block_timestamps[-last_n])
            else:
                timelength_seconds = (
                    block_timestamps[-1] - block_timestamps[-last_n]
                )
            timelength_clock = tooltime.timelength_to_clock(timelength_seconds)
            timelength_clock = timelength_clock.lstrip('0').lstrip(':')
            row.append(timelength_clock)
        else:
            row.append('last ' + last_times[index - len(last_as_int)])
            row.append(last_n)
            timelength_clock = tooltime.timelength_to_clock(
                last_times[index - len(last_as_int)]
            )
            timelength_clock = timelength_clock.lstrip('0').lstrip(':')
            row.append(timelength_clock)

        # add gas fee columns
        interval_median_gas_fees = median_gas_fees[-last_n:]
        if len(interval_median_gas_fees) > 0:
            row.append(np.nanmin(interval_median_gas_fees))
            row.append(np.nanmedian(interval_median_gas_fees))
            row.append(np.nanmean(interval_median_gas_fees))
            row.append(np.nanmax(interval_median_gas_fees))
        else:
            row.extend([None] * 4)

        rows.append(row)

    if export != 'stdout':
        import polars as pl

        final_df = pl.DataFrame(rows, labels, orient='row').sort('time')
        cli_utils.output_data(final_df, export, overwrite=overwrite)
    else:
        toolstr.print_table(
            rows=rows,
            labels=labels,
            border=styles['comment'],
            label_style=styles['title'],
            column_formats={
                'min': {'decimals': 1, 'trailing_zeros': True},
                'median': {'decimals': 1, 'trailing_zeros': True},
                'mean': {'decimals': 1, 'trailing_zeros': True},
                'max': {'decimals': 1, 'trailing_zeros': True},
            },
            column_styles={
                '': styles['metavar'],
                'blocks': styles['metavar'],
                'time': styles['metavar'],
                'min': styles['description'] + ' bold',
                'median': styles['description'] + ' bold',
                'mean': styles['description'] + ' bold',
                'max': styles['description'] + ' bold',
            },
        )

        xvals = block_timestamps
        yvals = median_gas_fees
        if using_latest:
            xtick_format = 'age'
            n_ticks = 3
        else:
            xtick_format = 'iso'
            n_ticks = 2
        plot = toolstr.render_line_plot(
            xvals=xvals,
            yvals=yvals,  # type: ignore
            n_rows=10,
            n_columns=60,
            line_style=styles['description'],
            chrome_style=styles['comment'],
            tick_label_style=styles['metavar'],
            xaxis_kwargs={
                'tick_label_format': xtick_format,
                'n_ticks': n_ticks,
            },
            char_dict=ctc.config.get_cli_chart_charset(),
        )
        print()
        print()
        toolstr.print(
            toolstr.hjustify('median fee', 'center', 70),
            indent=4,
            style=styles['title'],
        )
        toolstr.print(plot, indent=4)


async def async_print_gas_costs(
    *, gas_fee: float, eth_usd: float | None
) -> None:

    styles = cli.get_cli_styles()

    # see appendices G+H https://ethereum.github.io/yellowpaper/paper.pdf
    rows: typing.Sequence[typing.MutableSequence[typing.Any]] = [
        ['tx execution', 21000, 'gas per tx'],
        ['call data  (0)', 4, 'gas per byte'],
        ['call data (!0)', 16, 'gas per byte'],
        ['contract creation', 32000, 'gas per contract'],
        ['contract bytecode', 200, 'gas per byte'],
        ['contract destruct', -24000, 'gas per contract'],
        ['storage  0 --> !0', 20000 / 32, 'gas per byte'],
        ['storage  0 -->  0', 2900 / 32, 'gas per byte'],
        ['storage !0 -->  0', -15000 / 32, 'gas per byte'],
        ['log creation', 375, 'gas per log'],
        ['log topic', 375, 'gas per topic'],
        ['log data', 8, 'gas per byte'],
    ]
    labels = ['', 'gas cost', 'gas units']
    column_styles = [
        styles['metavar'],
        styles['description'] + ' bold',
        styles['option'],
    ]

    if eth_usd is not None:
        for row in rows:
            raw_gas_amount: int = row[1]
            gas_amount = float(raw_gas_amount)
            usd_cost = eth_usd * gas_fee * gas_amount / 1e9
            row.append(abs(usd_cost))
            row.append(row[2].replace('gas', '$'))
        labels.append('$ cost')
        labels.append('$ units')
        column_styles.append(styles['description'] + ' bold')
        column_styles.append(styles['option'])

    toolstr.print_table(
        rows,
        labels=labels,
        border=styles['comment'],
        label_style=styles['title'],
        column_styles=column_styles,
        column_formats={
            '$ cost': {
                'decimals': 5,
                'trailing_zeros': True,
                # 'scientific': False,
                'prefix': '$',
            },
        },
        column_gap=1,
        # compact=2,
        # indent=4,
    )
    print()
    toolstr.print(
        'for complete gas accounting see yellowpaper:\n    https://ethereum.github.io/yellowpaper/paper.pdf',
        style=styles['comment'],
        indent=4,
    )



from __future__ import annotations

import typing

from ctc import evm
from ctc import spec
from .. import rpc_request


def construct_eth_call(
    to_address: spec.BinaryData,
    *,
    from_address: spec.BinaryData | None = None,
    gas: spec.BinaryData | None = None,
    gas_price: spec.BinaryData | None = None,
    value_sent: spec.BinaryData | None = None,
    block_number: spec.BlockNumberReference | None = None,
    call_data: spec.BinaryData | None = None,
    function_parameters: typing.Sequence[typing.Any]
    | typing.Mapping[str, typing.Any]
    | None = None,
    function_abi: spec.FunctionABI | None = None,
) -> spec.RpcSingularRequest:

    if block_number is None:
        block_number = 'latest'
    else:
        block_number = evm.encode_block_number(block_number)

    # encode call data
    if call_data is None:
        call_data = evm.encode_call_data(
            parameters=function_parameters,
            function_abi=function_abi,
        )

    # assemble request data
    call_object = {
        'to': to_address,
        'data': call_data,
        'from': from_address,
        'gas': gas,
        'gasPrice': gas_price,
        'value': value_sent,
    }
    call_object = {k: v for k, v in call_object.items() if v is not None}

    return rpc_request.create('eth_call', [call_object, block_number])


def construct_eth_estimate_gas(
    to_address: spec.BinaryData,
    *,
    from_address: spec.BinaryData | None = None,
    gas: spec.BinaryData | None = None,
    gas_price: spec.BinaryData | None = None,
    value_sent: spec.BinaryData | None = None,
    call_data: spec.BinaryData | None = None,
    function_parameters: typing.Sequence[typing.Any]
    | typing.Mapping[str, typing.Any]
    | None = None,
    function_abi: spec.FunctionABI | None = None,
) -> spec.RpcSingularRequest:

    # encode call data
    if call_data is None:
        call_data = evm.encode_call_data(
            parameters=function_parameters,
            function_abi=function_abi,
        )

    # assemble call data
    call_object = {
        'to': to_address,
        'data': call_data,
        'from': from_address,
        'gas': gas,
        'gasPrice': gas_price,
        'value': value_sent,
    }
    call_object = {k: v for k, v in call_object.items() if v is not None}

    return rpc_request.create('eth_estimateGas', [call_object])


def construct_eth_get_balance(
    address: spec.Address,
    *,
    block_number: spec.BlockNumberReference | None = None,
) -> spec.RpcSingularRequest:

    if block_number is None:
        block_number = 'latest'

    encoded_block_number = evm.encode_block_number(block_number)
    return rpc_request.create('eth_getBalance', [address, encoded_block_number])


def construct_eth_get_storage_at(
    address: spec.BinaryData,
    position: spec.BinaryData,
    *,
    block_number: spec.BlockNumberReference = 'latest',
) -> spec.RpcSingularRequest:

    position = evm.to_hex(position, keep_leading_0=False)
    encoded_block_number = evm.encode_block_number(block_number)
    return rpc_request.create(
        'eth_getStorageAt', [address, position, encoded_block_number]
    )


def construct_eth_get_code(
    address: spec.BinaryData,
    *,
    block_number: spec.BlockNumberReference = 'latest',
) -> spec.RpcSingularRequest:

    block_number = evm.encode_block_number(block_number)
    return rpc_request.create('eth_getCode', [address, block_number])




from __future__ import annotations

from ctc import db

from ..fourbyte_db import fourbyte_statements


async_query_local_function_signatures = db.wrap_selector_with_connection(
    fourbyte_statements.async_select_function_signatures,
    '4byte',
)

async_query_local_event_signatures = db.wrap_selector_with_connection(
    fourbyte_statements.async_select_event_signatures,
    '4byte',
)



from __future__ import annotations

import time

import tooltime
import toolstr

import ctc.config
from ctc import cli
from ctc import evm
from ctc import rpc
from ctc import spec
from ctc.protocols import chainlink_utils
from . import chainlink_feed_metadata
from . import chainlink_spec


async def async_print_feed_summary(
    feed: str,
    *,
    start_block: spec.BlockNumberReference | None = None,
    n_recent: int | None = None,
    verbose: bool = False,
    context: spec.Context = None,
) -> None:

    import asyncio

    styles = cli.get_cli_styles()

    feed_address = await chainlink_feed_metadata.async_resolve_feed_address(
        feed,
        context=context,
    )

    name_coroutine = rpc.async_eth_call(
        function_abi=chainlink_spec.feed_function_abis['description'],
        to_address=feed_address,
        context=context,
    )
    decimals_coroutine = rpc.async_eth_call(
        function_abi=chainlink_spec.feed_function_abis['decimals'],
        to_address=feed_address,
        context=context,
    )
    aggregator_coroutine = rpc.async_eth_call(
        function_abi=chainlink_spec.feed_function_abis['aggregator'],
        to_address=feed_address,
        context=context,
    )
    name_task = asyncio.create_task(name_coroutine)
    decimals_task = asyncio.create_task(decimals_coroutine)
    aggregator_task = asyncio.create_task(aggregator_coroutine)

    if start_block is None:
        if n_recent is None:
            n_recent = 20
        latest_block = await rpc.async_eth_block_number(context=context)
        start_block = latest_block - 6000 * n_recent
    else:
        n_recent = 999999999999
    data = await chainlink_utils.async_get_feed_data(
        feed_address,
        fields='full',
        start_block=start_block,
        context=context,
    )
    timestamp = data['timestamp'][-1]

    updates = []
    for i, row in enumerate(data.to_dicts()):
        timestamp = int(row['timestamp'])
        age = tooltime.timelength_to_phrase(int(time.time() - timestamp))
        age = ' '.join(age.split(' ')[:2]).strip(',')
        update = [
            toolstr.format(row['answer'], decimals=4, trailing_zeros=True),
            age,
            str(i),
            str(timestamp),
            tooltime.timestamp_to_iso(timestamp).replace('T', ' '),
        ]
        updates.append(update)

    name = await name_task
    decimals = await decimals_task
    aggregator = await aggregator_task

    title = 'Chainlink Feed Summary: ' + name
    toolstr.print_text_box(title, double=False, style=styles['title'])
    cli.print_bullet(key='name', value=name)
    cli.print_bullet(key='decimals', value=str(decimals))

    metadata = await chainlink_feed_metadata.async_get_feed_metadata(
        feed,
        context=context,
    )
    deviation = toolstr.format(
        float(metadata['deviation']) / 100,
        percentage=True,
    )
    cli.print_bullet(key='deviation threshold', value=deviation)
    cli.print_bullet(key='heartbeat', value=metadata['heartbeat'])
    cli.print_bullet(key='address', value=feed_address)
    cli.print_bullet(key='aggregator', value=aggregator)

    creation_block = await chainlink_feed_metadata.async_get_feed_first_block(
        feed,
        context=context,
    )
    creation_timestamp = await evm.async_get_block_timestamp(
        creation_block,
        context=context,
    )
    cli.print_bullet(key='feed creation block', value=str(creation_block))
    cli.print_bullet(
        key='feed creation timestamp',
        value=str(creation_timestamp),
    )
    cli.print_bullet(
        key='feed age',
        value=tooltime.get_age(creation_timestamp, 'TimelengthPhrase'),
    )

    if verbose:
        print()
        print('fetching aggregator history...')
        history = await chainlink_utils.async_get_feed_aggregator_history(
            feed_address,
            context=context,
        )
        print()
        toolstr.print_header('Aggregator History', style=styles['title'])
        history_blocks = sorted(history.values())
        raw_timestamps = await evm.async_get_block_timestamps(
            history_blocks, context=context
        )
        history_timestamps = dict(zip(history.keys(), raw_timestamps))
        history_rows = []
        for old_aggregator, block in history.items():
            age_seconds: str = tooltime.get_age(
                history_timestamps[old_aggregator],
                'TimelengthPhrase',
            )
            age = ' '.join(age_seconds.split(' ')[:2]).strip(',')
            history_row = [
                old_aggregator,
                str(block),
                age,
            ]
            history_rows.append(history_row)
        labels = ['address', 'block', 'age']
        toolstr.print_table(
            history_rows,
            labels=labels,
            indent='    ',
            label_style=styles['title'],
            column_styles={'address': styles['metavar']},
            border=styles['comment'],
        )

    print()
    print()
    toolstr.print_header('Recent Updates', style=styles['title'])
    print()
    labels = ['value', 'age', 'block', 'timestamp', 'time']

    toolstr.print_table(
        updates[-n_recent:][::-1],
        labels=labels,
        indent='    ',
        column_formats={'value': {'decimals': 5}},
        label_style=styles['title'],
        border=styles['comment'],
        limit_rows=21,
        column_styles={
            'value': styles['description'] + ' bold',
            'age': styles['description'],
            'block': styles['option'],
            'timestamp': styles['option'],
            'time': styles['option'],
        },
    )

    xvals = data['timestamp'].to_list()
    yvals = data['answer'].to_list()
    plot = toolstr.render_line_plot(
        xvals=xvals,
        yvals=yvals,
        n_rows=10,
        n_columns=60,
        line_style=styles['description'],
        chrome_style=styles['comment'],
        tick_label_style=styles['metavar'],
        xaxis_kwargs={'tick_label_format': 'age'},
        char_dict=ctc.config.get_cli_chart_charset(),
    )
    print()
    print()
    print()
    toolstr.print(
        toolstr.hjustify(name + ' feed over time', 'center', 70),
        indent=4,
        style=styles['title'],
    )
    toolstr.print(plot, indent=4)




from __future__ import annotations

import toolsql


blocks_schema: toolsql.DBSchemaShorthand = {
    'tables': {
        'blocks': {
            'columns': [
                {'name': 'number', 'type': 'Integer', 'primary': True},
                {'name': 'hash', 'type': 'Text', 'index': True},
                {'name': 'timestamp', 'type': 'Integer', 'index': True},
                {'name': 'miner', 'type': 'Text'},
                {'name': 'extra_data', 'type': 'Text'},
                {'name': 'base_fee_per_gas', 'type': 'Integer', 'nullable': True},  # post-eip1559
                {'name': 'gas_limit', 'type': 'Integer'},
                {'name': 'gas_used', 'type': 'Integer'},
            ],
        },
    },
}



from __future__ import annotations

import asyncio

from ctc import config
from ctc import evm
from ctc import spec
from . import safe_spec


async def async_get_all_safes(
    *,
    start_block: spec.BlockReference | None = None,
    end_block: spec.BlockReference | None = None,
    context: spec.Context = None,
) -> spec.DataFrame:

    import polars as pl

    safes__1_1, safes__1_3 = await asyncio.gather(
        async_get_all_safes__1_1(
            start_block=start_block, end_block=end_block, context=context
        ),
        async_get_all_safes__1_3(
            start_block=start_block, end_block=end_block, context=context
        ),
    )

    return pl.concat([safes__1_1, safes__1_3])


async def async_get_all_safes__1_1(
    factory: spec.Address | None = None,
    *,
    start_block: spec.BlockReference | None = None,
    end_block: spec.BlockReference | None = None,
    context: spec.Context | None = None,
) -> spec.DataFrame:

    import polars as pl

    chain_id = config.get_context_chain_id(context)
    if factory is None:
        if (
            chain_id not in safe_spec.deployments
            or 'factory__1.1.1' not in safe_spec.deployments[chain_id]
        ):
            raise NotImplementedError(
                'must manually specify factory addresses, see '
                + safe_spec.safe_deployments_repository
            )
        else:
            factory = safe_spec.deployments[chain_id]['factory__1.1.1']

    df = await evm.async_get_events(
        factory,
        event_name='ProxyCreation',
        start_block=start_block,
        end_block=end_block,
        context=context,
    )
    old_columns = [
        'arg__proxy',
        'contract_address',
        'transaction_hash',
        'block_number',
    ]
    df = df[old_columns]

    if chain_id not in safe_spec.deployments:
        raise Exception('safe 1.1.1 deployment unknown for ' + str(chain_id))
    df = df.with_columns(
        pl.lit(safe_spec.deployments[chain_id]['safe__1.1.1']).alias('implementation'),
        pl.lit('1.1').alias('version'),
    )
    new_columns = {
        'arg__proxy': 'address',
        'contract_address': 'factory',
        'block_number': 'creation_block',
        'transaction_hash': 'creation_transaction',
    }
    df = df.rename(new_columns)
    return df


async def async_get_all_safes__1_3(
    factory: spec.Address | None = None,
    *,
    start_block: spec.BlockReference | None = None,
    end_block: spec.BlockReference | None = None,
    context: spec.Context = None,
) -> spec.DataFrame:

    import polars as pl

    if factory is None:
        chain_id = config.get_context_chain_id(context)
        if (
            chain_id not in safe_spec.deployments
            or 'factory__1.3.0' not in safe_spec.deployments[chain_id]
        ):
            raise NotImplementedError(
                'must manually specify factory addresses, see '
                + safe_spec.safe_deployments_repository
            )
        else:
            factory = safe_spec.deployments[chain_id]['factory__1.3.0']

    df = await evm.async_get_events(
        factory,
        event_name='ProxyCreation',
        start_block=start_block,
        end_block=end_block,
        context=context,
    )
    old_columns = [
        'arg__proxy',
        'contract_address',
        'transaction_hash',
        'block_number',
        'arg__singleton',
    ]
    df = df[old_columns]
    df = df.with_columns(
        pl.lit('1.3').alias('version'),
    )
    new_columns = {
        'arg__proxy': 'address',
        'contract_address': 'factory',
        'block_number': 'creation_block',
        'transaction_hash': 'creation_transaction',
        'arg__singleton': 'implementation',
    }
    df = df.rename(new_columns)
    return df




"""
the implementations here are stopgap solutions
- in future will create a more complete internal Uniswap v3 representation
"""
from __future__ import annotations

import typing

from ctc.toolbox import optimize_utils
from ctc import evm
from ctc import spec

from . import contracts


async def async_get_liquidity_depth(
    *,
    new_price: int | float,
    token_in: str,
    token_out: str,
    fee: int,
    min_search_depth: int = 1,
    max_search_depth: int = int(10e6 * 1e18),
    verbose: bool = False,
    output_tol: int | float | None = None,
    input_tol: int | float | None = None,
    max_iterations: int = 50,
    token_in_decimals: int | None = None,
    token_out_decimals: int | None = None,
    context: spec.Context = None,
) -> float:
    """return amount of token sold needed to reach new price"""

    if token_in_decimals is None:
        token_in_decimals = await evm.async_get_erc20_decimals(
            token_in, context=context
        )
    if token_out_decimals is None:
        token_out_decimals = await evm.async_get_erc20_decimals(
            token_out, context=context
        )

    if input_tol is None:
        input_tol = 10 ** (token_in_decimals - 2)

    swap_kwargs = {
        'token_in': token_in,
        'token_out': token_out,
        'fee': fee,
        'token_in_decimals': token_in_decimals,
        'token_out_decimals': token_out_decimals,
    }
    f_kwargs = {
        'target_new_price': new_price,
        'swap_kwargs': swap_kwargs,
    }

    try:
        return await optimize_utils.async_bisect(
            async_f=_async_new_price_distance,
            a=int(min_search_depth),
            b=int(max_search_depth),
            max_iterations=max_iterations,
            output_tol=output_tol,
            input_tol=input_tol,
            f_kwargs=f_kwargs,
            verbose=verbose,
        )
    except optimize_utils.BadSearchDomain:
        return 0


async def _async_new_price_distance(
    amount_sold: int | float,
    *,
    target_new_price: float,
    swap_kwargs: typing.Any,
    context: spec.Context = None,
) -> float:
    actual_new_price = await async_get_new_price(
        amount_sold=amount_sold, context=context, **swap_kwargs
    )
    return actual_new_price - target_new_price


async def async_get_new_price(
    *,
    amount_sold: int | float,
    token_in: str,
    token_out: str,
    fee: int,
    token_in_decimals: int,
    token_out_decimals: int,
    probe_amount_sold: int | None = None,
    context: spec.Context = None,
) -> float:
    """return new price in pool after a given amount of a token is sold

    - does this by computing a trade and an additional probe trade
    """
    if probe_amount_sold is None:
        probe_amount_sold = 10**token_in_decimals

    out_to_in = typing.cast(int, 10 ** (token_in_decimals - token_out_decimals))

    amount_sold = int(amount_sold)
    total_input_probe = int(amount_sold + probe_amount_sold)

    bought_amount = await contracts.async_quote_exact_input_single(
        token_in=token_in,
        token_out=token_out,
        fee=fee,
        amount_in=amount_sold,
        sqrt_price_limit_x96=0,
        context=context,
    )
    probe_amount_bought = await contracts.async_quote_exact_input_single(
        token_in=token_in,
        token_out=token_out,
        fee=fee,
        amount_in=total_input_probe,
        sqrt_price_limit_x96=0,
        context=context,
    )

    if not isinstance(bought_amount, int):
        raise Exception('invalid rpc result')
    if not isinstance(probe_amount_bought, int):
        raise Exception('invalid rpc result')

    probe_price = (
        out_to_in * (probe_amount_bought - bought_amount) / probe_amount_sold
    )

    return probe_price




from __future__ import annotations

from ctc import spec
from . import twap_data_sources
from . import twap_spec


async def async_get_data_feed(
    data_source: twap_spec.DataSource,
    *,
    start_block: int,
    end_block: int,
    context: spec.Context = None,
) -> spec.DataFrame:
    # acquire data
    protocol = data_source['protocol']
    if protocol == 'Chainlink':
        return await twap_data_sources.async_get_chainlink_data(
            start_block=start_block,
            end_block=end_block,
            data_source=data_source,
            context=context,
        )
    # elif protocol == 'UniswapV2':
    #     return await twap_data_sources.async_get_uniswap_v2_data(
    #         start_block=start_block,
    #         end_block=end_block,
    #         data_source=data_source,
    #         provider=provider,
    #     )
    else:
        raise Exception('unknown data source protocol: ' + str(protocol))



from __future__ import annotations

import typing

from ctc import spec
from .. import binary_utils

if typing.TYPE_CHECKING:
    from typing_extensions import TypeGuard


def is_address_str(some_str: typing.Any) -> TypeGuard[spec.Address]:
    """return whether input is an address str"""

    return (
        isinstance(some_str, str)
        and some_str.startswith('0x')
        and len(some_str) == 42
    )


def get_address_checksum(address: spec.Address) -> spec.Address:
    """return checksummed version of address str

    - adapted from eth_utils.to_checksum_address()
    """

    # validate address
    address_format = binary_utils.get_binary_format(address)
    if address_format not in ['prefix_hex', 'raw_hex']:
        raise Exception('checksum only relevant to hex formatted addresses')
    address = binary_utils.to_hex(address, prefix=False)

    # compute address hash
    address_hash = binary_utils.keccak_text(
        address.lower(),
        output_format='raw_hex',
    )

    # assemble checksum
    chars = []
    for address_char, hash_char in zip(address, address_hash):
        if int(hash_char, 16) > 7:
            chars.append(address_char.upper())
        else:
            chars.append(address_char.lower())
    raw_checksum = ''.join(chars)

    # convert to output format
    if address_format == 'raw_hex':
        return raw_checksum
    elif address_format == 'prefix_hex':
        return binary_utils.to_hex(raw_checksum)
    else:
        raise Exception('checksum only relevant to hex formatted addresses')



from __future__ import annotations

import typing

from ctc import rpc
from ctc import spec

from . import yearn_spec


async def async_get_tvl_adapter_assets(
    adapter: spec.Address,
    *,
    context: spec.Context = None
) -> typing.Sequence[spec.Address]:

    addresses: typing.Sequence[spec.Address] = await rpc.async_eth_call(
        to_address=adapter,
        function_name='assetsAddresses',
        n_parameters=0,
        context=context,
    )
    return addresses


async def async_get_tvl_adapter_assets_data(
    adapter: spec.Address,
    *,
    context: spec.Context = None,
) -> typing.Sequence[yearn_spec.AssetTvlBreakdown]:
    data = await rpc.async_eth_call(
        to_address=adapter,
        function_name='assetsTvlBreakdown',
        n_parameters=0,
        context=context,
    )

    tvls: typing.Sequence[yearn_spec.AssetTvlBreakdown] = [
        {
            'asset': datum[0],
            'token': datum[1],
            'token_price': datum[2],
            'underlying_balance': datum[3],
            'delegated_balance': datum[4],
            'adjusted_balance': datum[5],
            'tvl': datum[6],
        }
        for datum in data
    ]

    return tvls



import pytest

from ctc import evm


DAI_CONTRACT = "0x6B175474E89094C44Da98b954EedeAC495271d0F".lower()
START_BLOCK = 15181140
END_BLOCK = 15181147


@pytest.mark.asyncio
async def test_transfer_events():
    df = await evm.async_get_erc20_transfers(
        token=DAI_CONTRACT,
        start_block=START_BLOCK,
        end_block=END_BLOCK,
    )
    assert len(df) == 6
    assert df["contract_address"][0] == DAI_CONTRACT
    assert df["block_number"][0] == START_BLOCK
    assert "timestamp" not in df.columns



from __future__ import annotations

import toolsql


fourbyte_schema: toolsql.DBSchemaShorthand = {
    'tables': {
        'function_signatures': {
            'columns': [
                {'name': 'id', 'primary': True},
                {'name': 'created_at'},
                {'name': 'text_signature', 'index': True},
                {'name': 'hex_signature', 'index': True},
                {'name': 'bytes_signature'},
            ],
        },
        'event_signatures': {
            'columns': [
                {'name': 'id', 'primary': True},
                {'name': 'created_at'},
                {'name': 'text_signature', 'index': True},
                {'name': 'hex_signature', 'index': True},
                {'name': 'bytes_signature'},
            ],
        },
    },
}



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    import toolsql

from ctc import spec
from .. import config_values
from . import context_sources


def get_context_db_config(
    *,
    schema_name: str,
    context: spec.Context,
) -> toolsql.DBConfig:
    """get db_config of backend schema cache for a given context"""

    if not spec.is_schema_name(schema_name):
        raise Exception('not a valid schema name: ' + str(schema_name))
    backend = get_context_cache_backend(
        schema_name=schema_name, context=context
    )
    if backend is None:
        raise Exception(
            'no backend active for current context, cannot retrieve db_config'
        )
    return config_values.get_db_config(backend)


def get_context_cache_backend(
    *,
    schema_name: spec.db_types.SchemaName,
    context: spec.Context,
) -> str | None:
    """get backend schema cache for a given context"""
    backend, _read, _write = _get_context_cache_settings(
        schema_name=schema_name,
        context=context,
    )
    return backend


def get_context_cache_read_write(
    *,
    schema_name: str,
    context: spec.Context,
) -> tuple[bool, bool]:
    """get settings (read, write) of schema for a given context"""

    if not spec.is_schema_name(schema_name):
        raise Exception('not a valid schema name: ' + str(schema_name))
    _backend, read, write = _get_context_cache_settings(
        schema_name=schema_name,
        context=context,
    )

    return read, write


def _get_context_cache_settings(
    schema_name: spec.db_types.SchemaName,
    context: spec.Context,
) -> tuple[str | None, bool, bool]:
    """get settings (backend, read, write) of schema for a given context"""

    from ctc import config

    # get chain id if schema is network-related
    if schema_name in spec.network_schema_names:
        chain_id = context_sources.get_context_chain_id(context)
    else:
        chain_id = None

    # assemble rules from context and config
    config_rules = config.get_context_cache_rules()
    context_rules = _extract_context_cache_rules(context=context)
    rules = [rule for rl in [context_rules, config_rules] for rule in rl]

    return _resolve_context_cache_rules(
        rules=rules,
        chain_id=chain_id,
        schema_name=schema_name,
    )


def _extract_context_cache_rules(
    *,
    context_cache: spec.ContextCacheShorthand = None,
    context: spec.Context = None,
) -> typing.Sequence[spec.ContextCacheRule]:
    """convert a context's cache to a list of cache rules

    - this does not depend on the context_cache_rules of the config
    - this does depends on which networks, providers, or backends are in config
    """

    if context_cache is None:
        if isinstance(context, dict):
            context_cache = context.get('cache')
        else:
            return []

    if context_cache is None:
        return []
    elif isinstance(context_cache, bool):
        return [
            {'read': context_cache, 'write': context_cache},
        ]
    elif isinstance(context_cache, str):
        return [
            {'backend': context_cache},
        ]
    elif isinstance(context_cache, dict):
        keys = set(context_cache.keys())

        if keys.issubset({'filter', 'read', 'write', 'backend'}):
            # case: non-nested cache
            return [context_cache]  # type: ignore

        else:
            # case: nested caches
            rules = []
            for key, value in context_cache.items():
                rule = _create_nested_cache_rule(key, value)  # type: ignore
                if rule is not None:
                    rules.append(rule)

            return rules

    elif isinstance(context_cache, list):
        return context_cache
    else:
        raise Exception('unknown format for context_cache')


def _create_nested_cache_rule(
    key: str | int,
    value: None | bool | str | spec.ContextCacheRule,
) -> spec.ContextCacheRule | None:
    """assemble a single cache rule from within a shorthand nested cache spec"""

    from ctc import config

    # determine rule filter
    if key in spec.schema_names:
        filter: spec.ContextCacheFilter = {'schema': key}  # type: ignore
    elif isinstance(key, int):
        filter = {'chain_id': key}
    elif key in config.get_cache_backends():
        filter = {'backend': key}
    else:
        for chain_id, network_metadata in config.get_config_networks().items():
            if network_metadata['name'] == key:
                filter = {'chain_id': chain_id}
                break
        else:
            raise Exception('unknown cache identifier: ' + str(key))

    # determine rule settings
    if value is None:
        return None
    elif isinstance(value, bool):
        return {'filter': filter, 'read': value, 'write': value}
    elif isinstance(value, str):
        return {'filter': filter, 'backend': value}
    elif isinstance(value, dict):
        if not set(value.keys()).issubset({'backend', 'read', 'write'}):
            raise Exception('invalid keys for nested cache config')
        rule = value.copy()
        rule['filter'] = filter
        return rule
    else:
        raise Exception('unknown value for cache: ' + str(value))


def _resolve_context_cache_rules(
    rules: typing.Sequence[spec.ContextCacheRule],
    *,
    chain_id: spec.ChainId | None,
    schema_name: spec.SchemaName,
) -> tuple[str | None, bool, bool]:
    """get settings (backend, read, write) of schema for a given rule list"""

    backend: str | None = None
    read: bool | None = None
    write: bool | None = None

    # perform first pass to determine backend
    for rule in rules:
        if _does_cache_rule_apply(
            rule=rule, chain_id=chain_id, schema_name=schema_name
        ):
            if rule.get('backend') is not None:
                backend = rule['backend']
                break
    if backend is None:
        raise Exception('could not determine backend for cache')

    # perform second pass to determine read+write settings
    for rule in rules:
        if _does_cache_rule_apply(
            rule=rule,
            chain_id=chain_id,
            schema_name=schema_name,
            backend=backend,
        ):
            if read is None and rule.get('read') is not None:
                read = rule['read']
            if write is None and rule.get('write') is not None:
                write = rule['write']
    if read is None:
        raise Exception('could not determine read setting for cache')
    if write is None:
        raise Exception('could not determine write setting for cache')

    return backend, read, write


def _does_cache_rule_apply(
    rule: spec.ContextCacheRule,
    *,
    chain_id: spec.ChainId | None,
    schema_name: spec.SchemaName,
    backend: str | None = None,
) -> bool:
    """return whether a given context cache rule applies to given inputs"""

    # filter according to scoping rules
    filter = rule.get('filter')
    if filter is None:
        return True
    if filter.get('chain_id') is not None and filter['chain_id'] != chain_id:
        return False
    if filter.get('schema') is not None and filter['schema'] != schema_name:
        return False
    if filter.get('backend') is not None and filter['backend'] != backend:
        return False

    return True




from __future__ import annotations

import os
import typing

import toolstr

import ctc.config
from ctc.config import config_defaults

from . import cli_output_utils

if typing.TYPE_CHECKING:
    import toolcli

    StyleThemeReference = typing.Union[str, toolcli.StyleTheme]


color_themes: typing.Mapping[str, toolcli.StyleTheme] = {
    'apparatus': {
        # https://github.com/sslivkoff/apparatus-color-scheme
        # https://github.com/arcticicestudio/nord-vim/blob/main/colors/nord.vim
        'title': 'bold #ce93f9',
        'metavar': '#8be9fd',
        'description': '#b9f29f',
        'content': '#f1fa8c',
        'option': '#64aaaa',
        'comment': '#6272a4',
    },
    'nord': {
        # https://www.nordtheme.com/docs/colors-and-palettes
        'title': 'bold #88c0d0',
        'metavar': '#e5e9f0',
        'description': '#8fbcbb',
        'content': '#81a1c1',
        'option': '#5e81ac',
        'comment': '#aaaaaa',
    },
    # 'solarized': {
    #     # https://en.wikipedia.org/wiki/Solarized#Colors
    #     'title': 'bold #cb4b16',
    #     'metavar': '#268bd2',
    #     'description': '#859900',
    #     'content': '#6c71c4',
    #     'option': '#2aa198',
    #     'comment': '#657b83',
    # },
    'neonwolf': {
        # https://github.com/h3xx/tig-colors-neonwolf/blob/flair/screenshots/sample.png
        'title': 'bold #afff00',
        'metavar': '#5ccef4',
        'description': '#00afff',
        'content': '#ff0000',
        'option': '#d700ff',
        'comment': '#b2b2b2',
    },
    'monokai': {
        'title': 'bold #f92672',
        'metavar': '#a6e22e',
        'option': '#66d9ef',
        'description': '#fc9867',
        'content': '#ae81ff',
        'comment': '#88846f',
        # 'title': 'bold #66d9ef',
        # 'metavar': '#a6e22e',
        # 'option': '#e87d3e',
        # 'description': '#f92672',
        # 'content': '#ae81ff',
        # 'comment': '',
    },
    # 'oceanic': {
    #     # https://github.com/mhartington/oceanic-next
    #     'title': 'bold #fac863',
    #     'metavar': '#fac863',
    #     'option': '#fac863',
    #     'description': '#99c794',
    #     'content': '#c594c5',
    #     'comment': '#c594c5',
    #     # 'comment': '#6090c0',
    #     # 'metavar': '#ec5f67',
    #     # palenight seems similar https://github.com/drewtempelmeyer/palenight.vim/blob/master/images/screenshot.png
    # },
    # tokyo night https://github.com/folke/tokyonight.nvim
    # sonokai https://github.com/sainnhe/sonokai
    # '': {
    #     'title': 'bold ',
    #     'metavar': '',
    #     'option': '',
    #     'description': '',
    #     'content': '',
    #     'comment': '',
    # },
    # '': {
    #     'title': 'bold ',
    #     'metavar': '',
    #     'option': '',
    #     'description': '',
    #     'content': '',
    #     'comment': '',
    # },
    # '': {
    #     'title': 'bold ',
    #     'metavar': '',
    #     'option': '',
    #     'description': '',
    #     'content': '',
    #     'comment': '',
    # },
    # light themes
    # ayu https://github.com/ayu-theme/ayu-vim
    # solarized-light https://github.com/romainl/flattened
}


def get_cli_styles(color: bool | None = None) -> toolcli.StyleTheme:

    # if in notebook, do not use styles
    color = not _is_jupyter_notebook()

    if color:
        config_theme = ctc.config.get_cli_color_theme()

        # ensure that all keys are present
        config_theme = {
            'title': config_theme.get('title', ''),
            'metavar': config_theme.get('metavar', ''),
            'description': config_theme.get('description', ''),
            'content': config_theme.get('content', ''),
            'option': config_theme.get('option', ''),
            'comment': config_theme.get('comment', ''),
        }

        return config_theme

    else:
        return {
            'title': '',
            'metavar': '',
            'description': '',
            'content': '',
            'option': '',
            'comment': '',
        }


def get_color_theme_name() -> str:
    config_theme = ctc.config.get_cli_color_theme()
    for other_name, other_theme in color_themes.items():
        if other_theme == config_theme:
            return other_name
    else:
        return 'custom'


def _is_jupyter_notebook() -> bool:
    """adapted from https://gist.github.com/thomasaarholt/e5e2da71ea3ee412616b27d364e3ae82"""
    import sys

    if 'jupyter_client' not in sys.modules:
        return False

    try:
        from IPython import get_ipython

        if 'IPKernelApp' not in get_ipython().config:  # type: ignore
            raise ImportError("console")
            return False
        if 'VSCODE_PID' in os.environ:
            raise ImportError("vscode")
            return False
    except Exception:
        return False
    else:
        return True


def print_color_theme(color_theme: toolcli.StyleTheme) -> None:
    import toolstr

    for key, value in color_theme.items():
        cli_output_utils.print_bullet(
            key=key,
            value=toolstr.add_style(str(value), str(value)),
            key_style='',
        )


def set_color_theme(theme: StyleThemeReference) -> toolcli.StyleTheme:

    import json

    styles = resolve_color_theme_reference(theme)
    config_path = ctc.config.get_config_path()
    with open(config_path, 'r') as f:
        config_data = json.load(f)
    new_config_data = dict(config_data, cli_color_theme=styles)
    with open(config_path, 'w') as f:
        json.dump(new_config_data, f)
    print('color theme updated')

    return styles


def resolve_color_theme_reference(
    theme: StyleThemeReference | toolcli.StyleTheme,
) -> toolcli.StyleTheme:
    if isinstance(theme, str):
        if theme in ['default', 'reset']:
            styles = config_defaults.get_default_cli_color_theme()
        elif theme in color_themes:
            styles = color_themes[theme]
        else:
            raise Exception('unknown color theme: ' + str(theme))
    elif isinstance(theme, dict):
        styles = theme
    else:
        raise Exception('unknown theme format: ' + str(theme))

    return styles


def reset_color_theme() -> toolcli.StyleTheme:
    return set_color_theme('default')


def create_custom_color_theme(
    old_color_theme: toolcli.StyleTheme | None = None,
    headless: bool = False,
) -> toolcli.StyleTheme:

    import toolcli
    import toolstr

    if old_color_theme is None:
        old_color_theme = {
            'title': '',
            'metavar': '',
            'description': '',
            'content': '',
            'option': '',
            'comment': '',
        }

    new_color_theme: toolcli.StyleTheme = {
        'title': '',
        'metavar': '',
        'description': '',
        'content': '',
        'option': '',
        'comment': '',
    }
    for key, value in old_color_theme.items():
        style = toolcli.input_prompt(
            'What style to use for ' + toolstr.add_style(key, 'bold') + '? ',
            style=old_color_theme.get('metavar'),
            default='"' + toolstr.add_style(str(value), str(value)) + '"',
            headless=headless,
        )
        new_color_theme[key] = style  # type: ignore

    return new_color_theme


def preview_color_themes(
    themes: typing.Sequence[StyleThemeReference]
    | typing.Mapping[str, StyleThemeReference]
    | None = None,
) -> None:

    if themes is None:
        themes = color_themes

    theme_names = []
    theme_list = []
    if isinstance(themes, list):
        theme_names = [str(i + 1) for i in range(len(themes))]
        theme_list = [resolve_color_theme_reference(theme) for theme in themes]
    elif isinstance(themes, dict):
        theme_names = list(themes.keys())
        theme_list = [
            resolve_color_theme_reference(theme) for theme in themes.values()
        ]
    else:
        raise Exception()

    rows = []
    for theme_name, theme in zip(theme_names, theme_list):
        styled_names = []
        styled_styles = []
        for style_name, obj_style in theme.items():
            style = typing.cast(str, obj_style)
            styled_names.append(toolstr.add_style(style_name, style))
            styled_styles.append(toolstr.add_style(style, style))
        styles = '\n'.join(styled_names)
        row = [theme_name, styles, styled_styles]
        rows.append(row)
    labels = ['theme', 'labels', 'colors']
    toolstr.print_multiline_table(
        rows,
        labels=labels,
        label_style='bold',
        border='#666666',
    )



from __future__ import annotations


import toolcli

from ctc.protocols import curve_utils


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_curve_pools_command,
        'help': 'list curve pools',
        'args': [
            {
                'name': ['--verbose', '-v'],
                'help': 'include extra data',
                'action': 'store_true',
            },
        ],
        'examples': ['', '--verbose'],
    }


async def async_curve_pools_command(verbose: bool) -> None:
    import asyncio

    factories = [
        '0xb9fc157394af804a3578134a6585c0dc9cc990d4',
        '0x90e00ace148ca3b23ac1bc8c240c2a7dd9c2d7f5',
        '0x0959158b6040d32d04c301a72cbfd6b39e21c9ae',
        '0x8f942c20d02befc377d41445793068908e2250d0',
        '0xf18056bbd320e96a48e3fbf8bc061322531aac99',
    ]

    # get data from each factory
    coroutines = [
        curve_utils.async_get_factory_pool_data(factory, include_balances=False)
        for factory in factories
    ]
    factories_data = await asyncio.gather(*coroutines)
    items = [item for factory_data in factories_data for item in factory_data]

    # print as table
    completed = set()
    items = sorted(items, key=lambda item: ', '.join(item['symbols']))
    for item in items:

        if item['address'] in completed:
            continue
        else:
            completed.add(item['address'])

        if not verbose:
            skip = False
            for symbol in item['symbols']:
                if symbol.startswith('RC_'):
                    skip = True
            if skip:
                continue

        print(item['address'] + '    ' + ', '.join(item['symbols']))



from __future__ import annotations

import asyncio
import typing
import types

import tooltime
import polars as pl

from ctc import evm
from ctc import spec


async def async_get_lending_flows(
    wallet: spec.Address,
    pool_token: spec.ERC20Reference,
    *,
    protocol: typing.Literal['aave', 'compound', 'rari'],
    wallet_deposits: spec.DataFrame | None = None,
    deposits: spec.DataFrame | None = None,
    wallet_withdrawals: spec.DataFrame | None = None,
    withdrawals: spec.DataFrame | None = None,
    include_latest: bool = True,
    context: spec.Context = None,
    replace_symbols: bool = True,
    normalize: bool = True,
    include_rewards: bool = True,
) -> spec.DataFrame:

    if protocol == 'aave':
        from ctc.protocols import aave_v2_utils

        protocol_module: types.ModuleType = aave_v2_utils
    elif protocol == 'compound':
        from ctc.protocols import compound_utils

        protocol_module = compound_utils
    elif protocol == 'rari':
        from ctc.protocols import rari_utils

        protocol_module = rari_utils
    else:
        raise Exception('unknown protocol: ' + str(protocol))

    df = await _async_create_raw_wallet_flows_df(
        wallet=wallet,
        wallet_deposits=wallet_deposits,
        deposits=deposits,
        wallet_withdrawals=wallet_withdrawals,
        withdrawals=withdrawals,
        include_latest=include_latest,
        context=context,
    )

    underlying = await protocol_module.async_get_underlying_asset(
        pool_token=pool_token,
        context=context,
    )

    # add time data
    blocks = df['block_number'].to_numpy()
    blocks_before = list(blocks - 1)
    blocks_list = list(blocks)

    # queue tasks
    timestamps_coroutine = evm.async_get_block_timestamps(
        blocks=blocks_list,
        context=context,
    )
    timestamps_task = asyncio.create_task(timestamps_coroutine)
    pool_token_balances_before_coroutine = evm.async_get_erc20_balance_by_block(
        token=pool_token,
        wallet=wallet,
        blocks=blocks_before,
        context=context,
    )
    pool_token_balances_before_task = asyncio.create_task(
        pool_token_balances_before_coroutine
    )
    pool_token_balances_after_coroutine = evm.async_get_erc20_balance_by_block(
        token=pool_token,
        wallet=wallet,
        blocks=blocks_list,
        context=context,
    )
    pool_token_balances_after_task = asyncio.create_task(
        pool_token_balances_after_coroutine
    )
    asset_prices_coroutine = protocol_module.async_get_asset_price_by_block(
        asset=underlying,
        blocks=blocks_list,
        context=context,
    )
    asset_prices_task = asyncio.create_task(asset_prices_coroutine)

    # queue optional tasks
    if include_rewards:
        reward_coroutine = protocol_module.async_compute_wallet_rewards(
            wallet=wallet,
            blocks=blocks,
            context=context,
            replace_symbol=replace_symbols,
        )
        reward_task = asyncio.create_task(reward_coroutine)
    if normalize:
        decimals_coroutine = evm.async_get_erc20_decimals(
            underlying,
            context=context,
        )
        decimals_task = asyncio.create_task(decimals_coroutine)
    if replace_symbols:
        underlying_symbol_coroutine = evm.async_get_erc20_symbol(
            underlying,
            context=context,
        )
        underlying_symbol_task = asyncio.create_task(
            underlying_symbol_coroutine
        )
        pool_token_coroutine = evm.async_get_erc20_symbol(
            pool_token,
            context=context,
        )
        pool_token_symbol_task = asyncio.create_task(pool_token_coroutine)

    # normalize deposits and withdrawals
    if normalize:
        decimals = await decimals_task
        df['asset_deposit'] /= 10**decimals
        df['asset_withdrawal'] /= 10**decimals

    # compute time columns
    timestamps = await timestamps_task
    df = df.insert_at_idx(0, pl.Series('timestamp', timestamps))
    df = df.insert_at_idx(
        1,
        pl.Series('time', df['timestamp'].apply(tooltime.timestamp_to_iso)),
    )

    # add pool token balances
    df['pool_token_balance_before'] = await pool_token_balances_before_task
    df['pool_token_balance_after'] = await pool_token_balances_after_task

    # add underlying balances
    df['asset_balance_before'] = df['pool_token_balance_before']
    df['asset_balance_after'] = df['pool_token_balance_after']

    # add asset price
    df['asset_price'] = await asset_prices_task
    df['asset_balance_usd'] = df['asset_balance_after'] * df['asset_price']

    # add rewards
    rewards = await reward_task
    for key, value in rewards.items():
        df[key] = value

    # replace symbols
    if replace_symbols:
        rename_columns = {}
        underlying_symbol = await underlying_symbol_task
        pool_token_symbol = await pool_token_symbol_task
        for column in df.columns:
            if 'asset' in column:
                rename_columns[column] = column.replace(
                    'asset', underlying_symbol
                )
            if 'pool_token' in column:
                rename_columns[column] = column.replace(
                    'pool_token', pool_token_symbol
                )
        df = df.rename(rename_columns)

    return df


async def _async_create_raw_wallet_flows_df(
    *,
    wallet: spec.Address,
    wallet_deposits: spec.DataFrame | None = None,
    deposits: spec.DataFrame | None = None,
    wallet_withdrawals: spec.DataFrame | None = None,
    withdrawals: spec.DataFrame | None = None,
    include_latest: bool = True,
    context: spec.Context = None,
) -> spec.DataFrame:

    from ctc.protocols import aave_v2_utils

    no_deposits = wallet_deposits is None and deposits is None
    no_withdrawals = wallet_withdrawals is None and withdrawals is None
    if no_deposits and not no_withdrawals:
        deposits = await aave_v2_utils.async_get_deposits()
    elif not no_deposits and no_withdrawals:
        withdrawals = await aave_v2_utils.async_get_withdrawals()
    elif no_deposits and no_withdrawals:
        deposits, withdrawals = await asyncio.gather(
            aave_v2_utils.async_get_deposits(context=context),
            aave_v2_utils.async_get_withdrawals(context=context),
        )

    wallet = wallet.lower()
    if wallet_deposits is None:
        if deposits is None:
            raise Exception('could not determine deposits')
        wallet_deposits = deposits[deposits['arg__user'] == wallet]
    wallet_deposits = wallet_deposits.groupby('block_number').agg(pl.sum('*'))
    if spec.is_polars_dataframe(wallet_deposits):
        wallet_deposits_series = wallet_deposits['arg__amount']
    if wallet_withdrawals is None:
        if withdrawals is None:
            raise Exception('could not determine withdrawals')
        wallet_withdrawals = withdrawals.filter(withdrawals['arg__user'] == wallet)
    wallet_withdrawals = wallet_withdrawals.groupby('block_number').agg(pl.sum('*'))
    if spec.is_polars_dataframe(wallet_withdrawals):
        wallet_withdrawals_series = wallet_withdrawals['arg__amount']

    raw_data = {
        'block_number': wallet_withdrawals['block_number'],
        'asset_deposit': wallet_deposits_series,
        'asset_withdrawal': wallet_withdrawals_series,
    }
    if include_latest:
        raise NotImplementedError()
        # block = await evm.async_get_latest_block_number(context=context)
        # raw_data['block_number'].insert(0, block)
        # raw_data['asset_deposit'].insert(0, 0)
        # raw_data['asset_withdrawal'].insert(0, 0)
    raw_df = pl.DataFrame(raw_data)
    raw_df = raw_df.fill_null(0)

    return raw_df




from __future__ import annotations

import typing

from ctc import spec


def extract_bytecode_function_selectors(bytecode: str) -> typing.Sequence[str]:
    """extract solidity-style function selectors from contract bytecode"""
    import re

    # solidity style signatures
    results = re.findall('8063([a-f0-9]{8})146', bytecode)

    # vyper style signatures
    if len(results) == 0:
        results = re.findall('5[b,2]63([0-9a-f]{8})600051141561', bytecode)

    return ['0x' + result for result in results]


async def async_decompile_function_abis(
    bytecode: str,
    *,
    sort: str | None = None,
    context: spec.Context = None,
) -> typing.Sequence[typing.Mapping[str, typing.Any]]:
    """decompile solidity-style function ABI's from contract bytecode"""

    from ctc.protocols import fourbyte_utils

    function_selectors = extract_bytecode_function_selectors(bytecode)

    coroutines = [
        fourbyte_utils.async_query_function_signatures(
            selector, context=context
        )
        for selector in function_selectors
    ]

    import asyncio

    abi_lists = await asyncio.gather(*coroutines)
    abis: list[fourbyte_utils.Entry] = [
        abi for abi_list in abi_lists for abi in abi_list
    ]

    if sort is not None:
        abis = sorted(abis, key=lambda item: item[sort])  # type: ignore

    return abis




from __future__ import annotations

import ast
import typing

from ctc import evm
from ctc import spec
from . import safe_spec


async def async_get_safe_setup(
    safe_address: spec.Address,
    *,
    context: spec.Context = None,
) -> typing.Mapping[str, typing.Any] | None:
    creation_block = await evm.async_get_contract_creation_block(safe_address, context=context)
    events = await evm.async_get_events(
        safe_address,
        event_abi=safe_spec.event_abis['SafeSetup'],
        verbose=False,
        start_block=creation_block,
        end_block=creation_block,
        context=context,
    )
    if len(events) == 0:
        return None
    setup = events[:1].to_dicts()[0]
    setup['arg__owners'] = ast.literal_eval(setup['arg__owners'])
    return setup


async def async_get_safe_executions(
    safe_address: spec.Address,
    *,
    start_block: spec.BlockReference | None = None,
    end_block: spec.BlockReference | None = None,
    context: spec.Context = None,
) -> spec.DataFrame:
    return await evm.async_get_events(
        safe_address,
        event_abi=safe_spec.event_abis['ExecutionSuccess'],
        start_block=start_block,
        end_block=end_block,
        verbose=False,
        context=context,
    )


async def async_get_safe_owner_adds(
    safe_address: spec.Address,
    *,
    start_block: spec.BlockReference | None = None,
    end_block: spec.BlockReference | None = None,
    context: spec.Context = None,
) -> spec.DataFrame:
    return await evm.async_get_events(
        safe_address,
        event_abi=safe_spec.event_abis['AddedOwner'],
        start_block=start_block,
        end_block=end_block,
        verbose=False,
        context=context,
    )


async def async_get_safe_owner_removes(
    safe_address: spec.Address,
    *,
    start_block: spec.BlockReference | None = None,
    end_block: spec.BlockReference | None = None,
    context: spec.Context = None,
) -> spec.DataFrame:
    return await evm.async_get_events(
        safe_address,
        event_abi=safe_spec.event_abis['RemovedOwner'],
        start_block=start_block,
        end_block=end_block,
        verbose=False,
        context=context,
    )


async def async_get_safe_threshold_changes(
    safe_address: spec.Address,
    *,
    start_block: spec.BlockReference | None = None,
    end_block: spec.BlockReference | None = None,
    context: spec.Context = None,
) -> spec.DataFrame:
    return await evm.async_get_events(
        safe_address,
        event_abi=safe_spec.event_abis['ChangedThreshold'],
        start_block=start_block,
        end_block=end_block,
        verbose=False,
        context=context,
    )


async def async_get_safe_guard_changes(
    safe_address: spec.Address,
    *,
    context: spec.Context = None,
) -> spec.DataFrame:
    return await evm.async_get_events(
        safe_address,
        event_abi=safe_spec.event_abis['ChangedGuard'],
        verbose=False,
        context=context,
    )




from __future__ import annotations

import typing

import toolsql

from ctc import spec
from ... import schema_utils


async def async_upsert_contract_creation_block(
    *,
    address: spec.Address,
    block_number: int,
    context: spec.Context = None,
    conn: toolsql.AsyncConnection,
) -> None:

    table = schema_utils.get_table_schema(
        'contract_creation_blocks', context=context
    )
    await toolsql.async_insert(
        conn=conn,
        table=table,
        row={
            'address': address.lower(),
            'block_number': block_number,
        },
        upsert=True,
    )


async def async_select_contract_creation_block(
    address: spec.Address,
    *,
    context: spec.Context = None,
    conn: toolsql.AsyncConnection,
) -> int | None:

    table = schema_utils.get_table_schema(
        'contract_creation_blocks',
        context=context,
    )
    result = await toolsql.async_select(
        conn=conn,
        table=table,
        where_equals={'address': address.lower()},
        columns=['block_number'],
        output_format='cell_or_none',
    )

    if result is not None and not isinstance(result, int):
        raise Exception('invalid db result')

    return result


async def async_select_contract_creation_blocks(
    *,
    context: spec.Context = None,
    conn: toolsql.AsyncConnection,
) -> typing.Sequence[typing.Mapping[str, typing.Any]] | None:
    table = schema_utils.get_table_schema(
        'contract_creation_blocks',
        context=context,
    )
    result: typing.Sequence[typing.Mapping[str, typing.Any]] | None
    result = await toolsql.async_select(conn=conn, table=table)

    return result


async def async_delete_contract_creation_block(
    address: spec.Address,
    *,
    context: spec.Context = None,
    conn: toolsql.AsyncConnection,
) -> None:
    table = schema_utils.get_table_schema(
        'contract_creation_blocks',
        context=context,
    )
    await toolsql.async_delete(
        conn=conn,
        table=table,
        where_equals={'address': address.lower()},
    )




from __future__ import annotations

import typing

import toolcli
import toolstr

from ctc import cli
from ctc import evm
from ctc import spec


help_message = """decode EVM call data

syntax is one of
    1. [option]ctc decode call CALL_DATA[/option]
    2. [option]ctc decode call CONTRACT_ADDRESS CALL_DATA[/option]
    3. [option]ctc decode call TX_HASH --tx[/option]"""


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_decode_call_command,
        'help': help_message,
        'args': [
            {
                'name': 'args',
                'help': '`CALL_DATA` or `CONTRACT_ADDRESS CALL_DATA` or `TX_HASH`',
                'nargs': '+',
            },
            {
                'name': '--nested',
                'help': 'whether to decode nested calls',
                'action': 'store_true',
            },
        ],
        'examples': [
            '0xdac17f958d2ee523a2206206994597c13d831ec7 0xa9059cbb00000000000000000000000021dd5c13925407e5bcec3f27ab11a355a9dafbe3000000000000000000000000000000000000000000000000000000003fd629c0',
            '0xeefba1e63905ef1d7acba5a8513c70307c1ce441 0x252dba420000000000000000000000000000000000000000000000000000000000000020000000000000000000000000000000000000000000000000000000000000000200000000000000000000000000000000000000000000000000000000000000400000000000000000000000000000000000000000000000000000000000000100000000000000000000000000a65803ad604668e26a81be92c9f1c90354255eae00000000000000000000000000000000000000000000000000000000000000400000000000000000000000000000000000000000000000000000000000000044bba06f270000000000000000000000000a01960ddf19c59e43cbdf0b5ab9278d7459e76e000000000000000000000000e9822f18f2654e606a8dff9d75edd98367e7c0ae00000000000000000000000000000000000000000000000000000000000000000000000000000000a65803ad604668e26a81be92c9f1c90354255eae0000000000000000000000000000000000000000000000000000000000000040000000000000000000000000000000000000000000000000000000000000002451cff8d90000000000000000000000000a01960ddf19c59e43cbdf0b5ab9278d7459e76e00000000000000000000000000000000000000000000000000000000',
            '0xcb2931bd1c078f70106017f2e0c378fa9f648aa7d17c1d87240764b65e4626c9',
        ],
    }


async def async_decode_call_command(
    *,
    args: typing.Sequence[str],
    nested: bool = False,
    title: str | None = None,
    indent: str | None = None,
    mention_nested: bool = True,
    send_value: int | None = None,
    explicit_signature: str | None = None,
    context: spec.Context = None,
) -> None:

    tx = len(args) == 1 and (
        (args[0].startswith('0x') and len(args[0]) == 66)
        or (len(args[0]) == 64)
    )

    styles = cli.get_cli_styles()
    # print header
    if indent is None:
        indent = ''
    if title is None:
        title = 'Decoding Call Data'
    toolstr.print_text_box(title, style=styles['title'])

    if tx:
        if len(args) != 1:
            raise Exception('syntax is `ctc decode call TX_HASH --tx`')
        transaction = await evm.async_get_transaction(args[0], context=context)
        call_data = transaction['input']
        contract_address = transaction['to_address']
        contract_known = True
    elif len(args) == 1:
        from ctc.protocols import fourbyte_utils

        contract_known = False
        call_data = args[0]
        signature = call_data[:10]
        result = await fourbyte_utils.async_query_function_signatures(signature, context=context)

        for subresult in result:
            try:
                function_abi: spec.FunctionABI | None = (
                    evm.function_signature_to_abi(subresult['text_signature'])
                )
                decoded = evm.decode_call_data(
                    call_data=call_data,
                    function_abi=function_abi,
                )
                break
            except Exception as e:
                raise e
                continue
        else:
            raise Exception(
                'could not identify suitable abi for decoding function'
            )

    elif len(args) == 2:
        contract_address, call_data = args
        contract_known = True
    else:
        raise Exception('wrong syntax, see `ctc decode -h`')

    if contract_known:
        try:
            contract_abi = await evm.async_get_contract_abi(
                contract_address=contract_address,
                context=context,
            )
            if explicit_signature is not None:
                function_selector = evm.get_function_selector(
                    function_signature=explicit_signature
                )
                call_data = (
                    '0x'
                    + function_selector
                    + evm.to_hex(call_data, prefix=False)
                )

            decoded = evm.decode_call_data(
                contract_abi=contract_abi, call_data=call_data
            )
            function_abi = decoded['function_abi']
        except Exception:
            function_abi = None

    def print_bullet(
        key: typing.Any,
        value: typing.Any = '',
        indent: str = '',
        bullet: str = '-',
        key_style: str | None = None,
    ) -> None:
        if key_style is None:
            key_style = styles['option']

        as_str = indent + toolstr.add_style(bullet + ' ', styles['title'])
        if key != '':
            as_str += toolstr.add_style(
                str(key), key_style
            ) + toolstr.add_style(': ', styles['comment'])
        as_str += toolstr.add_style(str(value), styles['description'] + ' bold')
        toolstr.print(as_str)

    if contract_known:
        print_bullet('to', contract_address, indent)
    print_bullet(
        'n_bytes', len(evm.to_binary(call_data)), indent
    )
    if send_value is not None:
        print_bullet('value', send_value, indent)

    if function_abi is None:
        if explicit_signature is not None:
            print_bullet('signature', explicit_signature)
        print()
        if contract_known:
            print(
                '[abi for ' + contract_address + ' unavailable, cannot decode]'
            )
        else:
            print('[abi unavailable, cannot decode]')
        return

    # print funciton info
    print()
    print()
    toolstr.print_header('Function Info', style=styles['title'])
    print_bullet('name', function_abi['name'], indent)
    print_bullet('selector', decoded['function_selector'], indent)
    if explicit_signature is not None:
        print_bullet('signature', explicit_signature)
    else:
        print_bullet(
            'signature', evm.get_function_signature(function_abi), indent
        )
    print_bullet('inputs', indent=indent)
    for p, parameter in enumerate(function_abi['inputs']):
        if 'name' in parameter:
            cli.print_bullet(
                key=parameter['name'],
                value=parameter['type'],
                indent=indent + '    ',
                number=p + 1,
            )
        else:
            cli.print_bullet(
                value=parameter['type'],
                colon_str='',
                indent=indent + '    ',
                number=p + 1,
            )
    print_bullet('outputs', indent=indent)
    for p, parameter in enumerate(function_abi['outputs']):
        if 'name' in parameter:
            cli.print_bullet(
                key=parameter['name'],
                value=parameter['type'],
                indent=indent + '    ',
                number=(p + 1),
            )
        else:
            cli.print_bullet(
                key=parameter['name'],
                value=parameter['type'],
                indent=indent + '    ',
                number=(p + 1),
            )
    if len(function_abi['outputs']) == 0:
        toolstr.print(indent + '    \[no outputs]', style=styles['comment'])
    print()

    # detect nested calls
    nested_calls = []
    non_nested_parameters = list(range(len(decoded['parameters'])))
    named_parameters = decoded.get('named_parameters')

    function_selector = evm.get_function_selector(function_abi)
    if not nested:
        nested = function_selector in [
            '252dba42',  # aggregate((address,bytes)[])
            '7d5e81e2',  # propose(address[],uint256[],bytes[],string)
            '40a4f325',  # executeBatch(address,address[],uint256[],bytes[],bytes32,bytes32)
        ]

    if nested:

        if len(function_abi['inputs']) == 1:
            nested_calls = [
                {
                    'address': address,
                    'call_data': call_data,
                    'value': None,
                    'signature': None,
                }
                for address, call_data in decoded['parameters'][0]
            ]
            non_nested_parameters = []
        elif (
            named_parameters is not None
            and 'targets' in named_parameters
            and 'calldatas' in named_parameters
        ):

            # gather nested calls
            nested_calls = []
            for i in range(len(named_parameters['calldatas'])):
                nested_call = {
                    'address': named_parameters['targets'][i],
                    'call_data': named_parameters['calldatas'][i],
                    'value': None,
                    'signature': None,
                }
                if 'values' in named_parameters:
                    nested_call['value'] = named_parameters['values'][i]
                if 'signatures' in named_parameters:
                    if named_parameters['signatures'][i] != '':
                        nested_call['signature'] = named_parameters[
                            'signatures'
                        ][i]
                nested_calls.append(nested_call)

            # gather non-nested parameters
            non_nested_parameters = [
                n
                for n, name in enumerate(named_parameters)
                if name not in ['targets', 'calldatas', 'values', 'signatures']
            ]

        elif (
            named_parameters is not None
            and 'targets' in named_parameters
            and 'payloads' in named_parameters
            and 'values' in named_parameters
        ):
            nested_calls = [
                {
                    'address': target,
                    'call_data': payload,
                    'value': value,
                    'signature': None,
                }
                for target, payload, value in zip(
                    named_parameters['targets'],
                    named_parameters['payloads'],
                    named_parameters['values'],
                )
            ]

            # gather non-nested parameters
            non_nested_parameters = [
                n
                for n, name in enumerate(named_parameters)
                if name not in ['targets', 'payloads', 'values']
            ]

        else:
            raise Exception('could not detect nested calls')

    # print non nested parameters
    if len(non_nested_parameters) > 0:
        print()
        toolstr.print_header('Call Parameters', style=styles['title'])
        input_names = evm.get_function_parameter_names(function_abi)
        for p, parameter in enumerate(decoded['parameters']):
            if p not in non_nested_parameters:
                parameter = '\[nested parameter, see below]'  # type: ignore
            input_name = input_names[p]
            if input_name is None:
                input_name = function_abi['inputs'][p]['type']
            if isinstance(parameter, tuple):
                print_bullet(
                    str(input_name),
                    indent=indent,
                    bullet=str(p + 1) + '.',
                )
                for subparameter in parameter:
                    if isinstance(subparameter, bytes):
                        subparameter = evm.to_hex(subparameter)
                    toolstr.print(
                        indent + '    ' + stringify(subparameter),
                        style=styles['description'],
                    )
                if len(parameter) == 0:
                    toolstr.print(
                        indent + '    \[none]', style=styles['comment']
                    )
            else:
                if isinstance(parameter, bytes):
                    parameter = evm.to_hex(parameter)
                print_bullet(
                    input_name,
                    parameter,
                    indent=indent,
                    bullet=str(p + 1) + '.',
                )

    # print nested calls
    if nested:
        for nc, nested_call in enumerate(nested_calls):
            # nested_address, nested_call_data = nested_call
            title = (
                'Nested Call Data '
                + str(nc + 1)
                + ' / '
                + str(len(nested_calls))
            )
            print()
            await async_decode_call_command(
                args=[nested_call['address'], nested_call['call_data']],
                nested=False,
                title=title,
                indent='',
                mention_nested=False,
                send_value=nested_call['value'],
                explicit_signature=nested_call['signature'],
                context=context,
            )
            if nc + 1 != len(nested_calls):
                print()

    if mention_nested and not nested:
        toolstr.print(
            'use --nested to decode nested calls',
            style=styles['comment'],
        )


def stringify(item: typing.Any) -> str:
    return str(item)
    if isinstance(item, tuple):
        return str(list(stringify(subitem) for subitem in item))
    else:
        return str(item)



from .rpc_batch import *
from .rpc_constructors import *
from .rpc_digestors import *
from .rpc_executors_async import *
from .rpc_executors_sync import *
from .rpc_protocols import *

from .rpc_lifecycle import *
from .rpc_provider import *
from .rpc_registry import *
from .rpc_request import *
from .rpc_spec import *

# do not import
# - rpc_format
# - rpc_logging




from __future__ import annotations

from ctc import spec
from ... import binary_utils
from .. import function_abi_utils


def get_event_hash(event_data: spec.EventABI | str) -> str:
    """compute event hash from event signature or event abi"""
    if isinstance(event_data, str):
        signature = event_data
    else:
        signature = get_event_signature(event_abi=event_data)
    return binary_utils.keccak_text(signature)


def get_event_signature(event_abi: spec.EventABI) -> str:
    """get event signature from event ABI"""
    import eth_utils_lite  # type: ignore

    base_arg_types = [var['type'] for var in event_abi['inputs']]
    arg_types = []
    for i, item in enumerate(base_arg_types):
        if item.startswith('tuple'):
            arg_type = eth_utils_lite.abi.collapse_if_tuple(
                event_abi['inputs'][i]
            )
        else:
            arg_type = function_abi_utils.get_function_selector_type(item)
        arg_types.append(arg_type)
    inputs = ','.join(arg_types)
    return event_abi['name'] + '(' + inputs + ')'


def get_event_unindexed_types(
    event_abi: spec.EventABI,
) -> list[spec.ABIDatumType]:
    """get list of data types in signature of event"""

    import eth_utils_lite

    return [
        eth_utils_lite.abi.collapse_if_tuple(var)
        for var in event_abi['inputs']
        if not var['indexed']
    ]


def get_event_unindexed_names(event_abi: spec.EventABI) -> list[str]:
    """get list of data names in signature of event"""
    return [var['name'] for var in event_abi['inputs'] if not var['indexed']]


def get_event_indexed_names(event_abi: spec.EventABI) -> list[str]:
    """get list of indexed names in signature of event"""
    return [var['name'] for var in event_abi['inputs'] if var['indexed']]


def get_event_indexed_types(
    event_abi: spec.EventABI,
) -> list[spec.ABIDatumType]:
    """get list of indexed types in signature of event"""
    return [var['type'] for var in event_abi['inputs'] if var['indexed']]


def get_event_schema(event_abi: spec.EventABI) -> spec.EventSchema:
    """return schema of types and names of event"""
    indexed_names = get_event_indexed_names(event_abi)
    indexed_types = get_event_indexed_types(event_abi)
    unindexed_names = get_event_unindexed_names(event_abi)
    unindexed_types = get_event_unindexed_types(event_abi)
    return {
        'indexed_names': indexed_names,
        'indexed_types': indexed_types,
        'unindexed_names': unindexed_names,
        'unindexed_types': unindexed_types,
        'names': indexed_names + unindexed_names,
        'types': indexed_types + unindexed_types,
    }




from __future__ import annotations

import copy
import typing

import toolsql

from ctc import spec
from ... import schema_utils


def _format_dex_pool(dex_pool: spec.DexPool) -> spec.DexPool:
    formatted = copy.copy(dex_pool)
    return formatted


async def async_upsert_dex_pool(
    *,
    dex_pool: spec.DexPool,
    conn: toolsql.AsyncConnection,
    context: spec.Context,
) -> None:
    table = schema_utils.get_table_schema('dex_pools', context=context)
    dex_pool = _format_dex_pool(dex_pool)
    await toolsql.async_insert(
        conn=conn,
        table=table,
        row=dex_pool,
        upsert=True,
    )


async def async_upsert_dex_pools(
    *,
    dex_pools: typing.Sequence[spec.DexPool],
    conn: toolsql.AsyncConnection,
    context: spec.Context,
) -> None:
    if len(dex_pools) == 0:
        return
    table = schema_utils.get_table_schema('dex_pools', context=context)
    dex_pools = [_format_dex_pool(dex_pool) for dex_pool in dex_pools]
    await toolsql.async_insert(
        conn=conn,
        table=table,
        rows=dex_pools,
        upsert=True,
    )


async def async_upsert_dex_pool_factory_query(
    *,
    factory: spec.Address,
    last_scanned_block: int,
    conn: toolsql.AsyncConnection,
    context: spec.Context,
) -> None:
    table = schema_utils.get_table_schema(
        'dex_pool_factory_queries',
        context=context,
    )
    await toolsql.async_insert(
        conn=conn,
        table=table,
        row={
            'factory': factory.lower(),
            'last_scanned_block': last_scanned_block,
        },
        upsert=True,
    )


async def async_delete_dex_pool(
    *,
    conn: toolsql.AsyncConnection,
    dex_pool: spec.Address,
    context: spec.Context = None,
) -> None:

    table = schema_utils.get_table_schema('dex_pools', context=context)

    await toolsql.async_delete(
        conn=conn,
        table=table,
        where_equals={'address': dex_pool.lower()},
    )


async def async_delete_dex_pools(
    *,
    conn: toolsql.AsyncConnection,
    dex_pools: typing.Sequence[spec.Address],
    context: spec.Context = None,
) -> None:

    table = schema_utils.get_table_schema('dex_pools', context=context)

    await toolsql.async_delete(
        conn=conn,
        table=table,
        where_in={'address': [dex_pool.lower() for dex_pool in dex_pools]},
    )


async def async_delete_dex_pool_factory_query(
    *,
    conn: toolsql.AsyncConnection,
    factory: spec.Address,
    context: spec.Context = None,
) -> None:

    table = schema_utils.get_table_schema(
        'dex_pool_factory_queries', context=context
    )

    await toolsql.async_delete(
        conn=conn,
        table=table,
        where_equals={'factory': factory.lower()},
    )


async def async_select_dex_pool(
    address: spec.Address,
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> spec.DexPool | None:

    table = schema_utils.get_table_schema('dex_pools', context=context)

    where_equals = {'address': address.lower()}

    result = await toolsql.async_select(
        conn=conn,
        table=table,
        where_equals=where_equals,
        output_format='single_dict_or_none',
    )
    return result  # type: ignore


async def async_select_dex_pools_by_id(
    addresses: typing.Sequence[spec.Address],
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> typing.Mapping[spec.Address, spec.DexPool | None] | None:
    raise NotImplementedError()

    if len(addresses) == 0:
        return {}

    table = schema_utils.get_table_schema('dex_pools', context=context)

    results = await toolsql.async_select(
        conn=conn,
        table=table,
        where_in={'address': addresses},
    )

    results_by_address = {
        row['address']: row for row in results if row is not None
    }

    return {address: results_by_address.get(address) for address in addresses}


async def async_select_dex_pools(
    *,
    factory: spec.Address | None = None,
    factories: typing.Sequence[spec.Address] | None = None,
    assets: typing.Sequence[spec.Address] | None = None,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
    start_block: int | None = None,
    end_block: int | None = None,
) -> typing.Sequence[spec.DexPool] | None:

    table = schema_utils.get_table_schema('dex_pools', context=context)

    query: typing.MutableMapping[str, typing.Any] = {}
    if factory is not None:
        query.setdefault('where_equals', {})
        query['where_equals']['factory'] = factory
    if factories is not None:
        query.setdefault('where_in', {})
        query['where_in']['factory'] = factories
    if assets is not None:
        for asset in assets:
            asset = asset.lower()
            query['where_or'] = [
                {'where_equals': {'asset0': asset}},
                {'where_equals': {'asset1': asset}},
                {'where_equals': {'asset2': asset}},
                {'where_equals': {'asset3': asset}},
            ]
    if start_block is not None:
        query.setdefault('where_gte', {})
        query['where_gte']['creation_block'] = start_block
    if end_block is not None:
        query.setdefault('where_lte', {})
        query['where_lte']['creation_block'] = end_block

    return await toolsql.async_select(  # type: ignore
        conn=conn,
        table=table,
        **query,
    )


async def async_select_dex_pool_factory_last_scanned_block(
    factory: spec.Address,
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context,
) -> int | None:

    table = schema_utils.get_table_schema(
        'dex_pool_factory_queries', context=context
    )

    result = await toolsql.async_select(
        conn=conn,
        table=table,
        columns=['last_scanned_block'],
        output_format='cell_or_none',
        where_equals={'factory': factory.lower()},
    )

    if result is None or isinstance(result, int):
        return result
    else:
        raise Exception('bad data format received from db')




from __future__ import annotations

import os
import typing
from ctc import cli

if typing.TYPE_CHECKING:
    from ctc import spec


def print_bullet(
    value: typing.Any,
    *,
    key: typing.Any | None = None,
    number: int | None = None,
    colon_str: str | None = None,
    indent: int | str | None = None,
    bullet_style: str | None = None,
    colon_style: str | None = None,
    key_style: str | None = None,
    value_style: str | None = None,
) -> None:

    import toolstr

    styles = cli.get_cli_styles()

    if bullet_style is None:
        bullet_style = styles['title']
    if colon_style is None:
        colon_style = styles['comment']
    if key_style is None:
        key_style = styles['option']
    if value_style is None:
        value_style = styles['description']

    toolstr.print_bullet(
        key=key,
        value=value,
        number=number,
        colon_str=colon_str,
        indent=indent,
        bullet_style=bullet_style,
        colon_style=colon_style,
        key_style=key_style,
        value_style=value_style,
    )


def output_data(
    data: spec.DataFrame | spec.Series,
    output: str,
    *,
    overwrite: bool,
    top: int | None = None,
    indent: str | int | None = None,
    raw: bool = False,
) -> None:

    if output == 'stdout':
        import toolstr

        if top is not None:
            data = data[:top]
        toolstr.print_dataframe_as_table(data, indent=indent, )

    else:
        import toolcli
        import polars as pl

        # check whether file exists
        if os.path.isfile(output):
            if overwrite:
                pass
            elif toolcli.input_yes_or_no('File already exists. Overwrite? '):
                pass
            else:
                raise Exception('aborting')

        if spec.is_polars_series(data):
            df = pl.DataFrame(data)
        elif spec.is_polars_dataframe(data):
            df = data
        else:
            raise Exception('unknown data format')

        if output.endswith('.parquet'):
            df.write_parquet(output)
        elif output.endswith('.csv'):
            df.write_csv(output)
        elif output.endswith('.json'):
            df.write_json(output)
        else:
            raise Exception('unknown output format: ' + str(output))



from __future__ import annotations

import typing

import toolcli
import toolsql

import ctc.config


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_create_tables_command,
        'help': 'create tables for database',
        'args': [
            {
                'name': '--schema-names',
                'nargs': '+',
                'help': 'schemas to create tables for',
            },
            {
                'name': '--networks',
                'nargs': '+',
                'help': 'networks to create tables for',
            },
        ],
        'examples': [''],
        'hidden': True,
    }


async def async_create_tables_command(
    schema_names: typing.Sequence[str],
    networks: typing.Sequence[str],
) -> None:
    from ctc import db

    if schema_names is not None:
        for schema_name in schema_names:
            if schema_name not in db.DBSchemaName.__args__:  # type: ignore
                raise Exception('unknown schema_name: ' + str(schema_name))

    db_config = ctc.config.get_context_db_config(
        context=None, schema_name=schema_name
    )
    with toolsql.connect(db_config) as conn:
        db.create_missing_tables(
            schema_names=schema_names,
            networks=networks,
            conn=conn,
        )




from __future__ import annotations

import typing

import msgspec


class Log(msgspec.Struct, rename='camel'):
    block_number: str
    transaction_index: str
    log_index: str
    transaction_hash: str
    address: str
    topics: tuple[str, ...]
    data: str
    removed: bool
    block_hash: str


class RpcResult(msgspec.Struct):
    result: list[Log]


decoder = msgspec.json.Decoder(RpcResult)


def decode_logs(
    raw_response: str,
    include_removed: bool = False,
) -> typing.Sequence[tuple[int, int, int, str, str, tuple[str, ...], str, str]]:

    response = decoder.decode(raw_response)
    logs = response.result

    return [
        (
            int(log.block_number, 16),
            int(log.transaction_index, 16),
            int(log.log_index, 16),
            log.transaction_hash,
            log.address,
            log.topics,
            log.data,
            log.block_hash,
        )
        for log in logs
        if (not log.removed or include_removed)
    ]




from __future__ import annotations


def get_stable_version(version_str: str) -> str:
    if version_str.count('.') != 2:
        raise Exception('invalid version spec')
    major, minor, bugfix = version_str.split('.')
    if not major.isnumeric() or not minor.isnumeric():
        raise Exception('invalid version spec')
    for i in range(len(bugfix)):
        if not bugfix[i].isnumeric():
            stable_bugfix = bugfix[:i]
            break
    else:
        stable_bugfix = bugfix
    if len(stable_bugfix) == 0:
        raise Exception('invalid version spec')
    return '.'.join([major, minor, stable_bugfix])



from __future__ import annotations

import typing

from ctc import spec
from ... import binary_utils
from . import function_abi_queries


def function_signature_to_abi(function_signature: str) -> spec.FunctionABI:
    """return a partial ABI of function from just a signature

    will be missing input parameter names and information about outputs
    """
    function_name, tail = function_signature.split('(')
    parameters, _ = tail.split(')')
    parameter_types = parameters.split(',')

    return {
        'type': 'function',
        'name': function_name,
        'inputs': [
            {'type': parameter_type} for parameter_type in parameter_types
        ],
        'outputs': [],
    }


async def async_parse_function_str_abi(
    function: str,
    contract_address: spec.Address | None = None,
    *,
    context: spec.Context = None,
) -> spec.FunctionABI:
    """parse a function str into a function ABI

    function str can be a function name, 4byte selector, or ABI

    used for cli commands
    """

    if is_function_selector(function):
        # function given as a function selector
        function_name = None
        function_selector = function
        function_abi = None
    elif is_function_signature(function):
        return function_signature_to_abi(function)
    else:
        try:
            # function given as json / python dict str
            import ast

            function_name = None
            function_selector = None
            function_abi = ast.literal_eval(function)
        except Exception:
            # function given as a function name
            function_name = function
            function_selector = None
            function_abi = None

    if function_abi is None:
        function_abi = await function_abi_queries.async_get_function_abi(
            contract_address=contract_address,
            function_name=function_name,
            function_selector=function_selector,
            context=context,
        )

    return function_abi


def get_function_parameter_types(
    function_abi: spec.FunctionABI | None = None,
    function_signature: typing.Optional[str] = None,
) -> list[spec.ABIDatumType]:
    """return list of function parameter types"""

    if function_abi is not None:

        output = []
        for item in function_abi.get('inputs', []):
            import eth_utils_lite  # type: ignore

            cast_item = typing.cast(typing.Dict[str, typing.Any], item)
            collapsed = eth_utils_lite.abi.collapse_if_tuple(cast_item)
            output.append(collapsed)

        return output

    elif function_signature is not None:
        parameter_str = function_signature.split('(')[1]
        parameter_str = parameter_str[:-1]
        return parameter_str.split(',')

    else:
        raise Exception('must specify function_abi or function_signature')


@typing.overload
def get_function_parameter_names(
    function_abi: spec.FunctionABI,
    require_names: typing.Literal[True],
) -> list[str]:
    ...


@typing.overload
def get_function_parameter_names(
    function_abi: spec.FunctionABI,
    require_names: typing.Literal[False] = False,
) -> list[typing.Optional[str]]:
    ...


def get_function_parameter_names(
    function_abi: spec.FunctionABI,
    require_names: bool = False,
) -> typing.Union[list[str], list[typing.Optional[str]]]:
    """return list of function parameter names"""

    names = [item.get('name') for item in function_abi.get('inputs', [])]

    if require_names:
        if any(name is None for name in names):
            raise Exception('function abi does not specify names')
        else:
            return typing.cast(typing.List[str], names)
    else:
        return names


def get_function_signature(
    function_abi: spec.FunctionABI | None = None,
    *,
    parameter_types: typing.Optional[list[str]] = None,
    function_name: typing.Optional[str] = None,
    include_names: bool = False,
) -> str:
    """return function signature"""

    # get parameter types
    if parameter_types is None:
        parameter_types = get_function_parameter_types(function_abi)
    parameter_types = [
        get_function_selector_type(item) for item in parameter_types
    ]

    # create entires with or without names
    if include_names:
        if function_abi is None:
            raise Exception('must specify function_abi')
        parameter_names = get_function_parameter_names(
            function_abi, require_names=True
        )
        entries = [
            type + ' ' + name
            for type, name in zip(parameter_types, parameter_names)
        ]
        entries_str = ', '.join(entries)
    else:
        entries = parameter_types
        entries_str = ','.join(entries)

    if function_name is None:
        if function_abi is None:
            raise Exception('must specify function_abi')
        function_name = function_abi['name']

    return function_name + '(' + entries_str + ')'


def get_function_selector_type(
    datatype: spec.ABIDatumType,
) -> spec.ABIDatumType:
    """get the data name used for function selector"""
    if datatype == 'uint':
        return 'uint256'
    elif datatype == 'int':
        return 'int256'
    elif datatype == 'ufixed':
        return 'ufixed128x18'
    elif datatype == 'fixed':
        return 'fixed128x18'
    else:
        return datatype


def get_function_selector(
    function: spec.FunctionABI | spec.FunctionSignature | None = None,
    *,
    function_abi: typing.Optional[spec.FunctionABI] = None,
    function_signature: typing.Optional[spec.FunctionSignature] = None,
) -> str:
    """get function 4 byte selector

    TODO: deprecate function_abi and function_signature inputs
    """

    if isinstance(function, dict):
        function_abi = function
    elif isinstance(function, str):
        function_signature = function
    elif function is None:
        assert function_abi is not None or function_signature is not None
    else:
        raise Exception('unknown funciton format: ' + str(type(function)))

    if function_signature is None:
        if function_abi is None:
            raise Exception('must specify function_abi or function_signature')
        function_signature = get_function_signature(function_abi)

    full_hash = binary_utils.keccak(
        function_signature.encode(), output_format='raw_hex'
    )
    return full_hash[:8]


def is_function_selector(selector: typing.Any) -> bool:
    """return whether input is a function selector"""
    return isinstance(selector, str) and (
        (spec.is_prefix_hex_data(selector) and len(selector) == 10)
        or (spec.is_raw_hex_data(selector) and len(selector) == 8)
    )


def is_function_signature(signature: typing.Any) -> bool:
    """return whether a str is a function signature

    this is NOT a comprehensive test, only an approximation using regexes
    """
    if isinstance(signature, str):
        import re

        return re.fullmatch('[a-zA-Z_].*\([a-z,\[\]]*\)', signature) is not None
    return False


def get_function_output_types(
    function_abi: spec.FunctionABI,
) -> list[spec.ABIDatumType]:
    """return list of function output types"""

    import eth_utils_lite

    output_types = []
    for output in function_abi['outputs']:
        cast_output = typing.cast(typing.Dict[str, typing.Any], output)
        output_type = eth_utils_lite.abi.collapse_if_tuple(cast_output)
        output_types.append(output_type)
    return output_types


def get_function_output_names(
    function_abi: spec.FunctionABI,
    human_readable: bool = False,
) -> list[typing.Optional[str]]:
    """return list of function output names"""

    output_names = [output.get('name') for output in function_abi['outputs']]

    # human readable uses function name or output_# if no names provided
    if human_readable:
        if len(output_names) == 1:
            if output_names[0] == '':
                output_names[0] = function_abi.get('name')
        else:
            for on, output_name in list(enumerate(output_names)):
                if output_name == '':
                    output_names[on] == 'output_' + str(on)

    return output_names


def is_function_read_only(function_abi: spec.FunctionABI) -> bool:
    """return whether function is read-only"""
    return bool(function_abi.get('constant')) or (
        function_abi.get('stateMutability')
        in (
            'view',
            'pure',
        )
    )



"""
## Next type system for RPC requests/responses
- create RpcRequest[Method] and RpcResponse[Method]
- implement using mypy plugin with get_type_analyze_hook()
    - https://github.com/python/mypy/issues/6503
    - https://github.com/python/mypy/issues/8848
"""
from __future__ import annotations

import typing
from typing_extensions import TypedDict, Literal

from . import network_types


RpcSingularRequest = typing.Dict[str, typing.Any]
RpcPluralRequest = typing.List[RpcSingularRequest]
RpcRequest = typing.Union[RpcSingularRequest, RpcPluralRequest]


class RpcSingularResponseSuccess(TypedDict):
    id: str
    jsonrpc: str
    result: typing.Any


class RpcSingularResponseFailure(TypedDict):
    id: str
    jsonrpc: str
    error: typing.Any


RpcSingularResponseRaw = typing.Union[
    RpcSingularResponseSuccess,
    RpcSingularResponseFailure,
]


RpcPluralResponseRaw = typing.List[RpcSingularResponseRaw]
RpcResponseRaw = typing.Union[RpcSingularResponseRaw, RpcPluralResponseRaw]

RpcSingularResponse = typing.Any
RpcPluralResponse = typing.List[RpcSingularResponse]
RpcResponse = typing.Union[RpcPluralResponse, RpcSingularResponse]

RpcConstructor = typing.Callable[..., RpcSingularRequest]
RpcDigestor = typing.Callable[..., RpcResponse]


#
# # provider
#

ProviderShortcut = str


class PartialProvider(TypedDict, total=False):
    url: str
    name: str | None
    network: network_types.NetworkReference | None
    protocol: Literal['http', 'wss', 'ipc']
    #
    # query behaviors
    session_kwargs: typing.Mapping[str, typing.Any] | None
    chunk_size: int | None
    convert_reverts_to_none: bool
    disable_batch_requests: bool


class Provider(TypedDict, total=True):
    url: str
    name: str | None
    network: network_types.ChainId
    protocol: Literal['http', 'wss', 'ipc']
    #
    # query behaviors
    session_kwargs: typing.Mapping[str, typing.Any] | None
    chunk_size: int | None
    convert_reverts_to_none: bool
    disable_batch_requests: bool


ProviderReference = typing.Union[ProviderShortcut, PartialProvider, Provider]
ProviderId = typing.Tuple[
    int, str, typing.Tuple[typing.Tuple[typing.Any, typing.Any], ...]
]




from __future__ import annotations

import typing

from ctc import spec

if typing.TYPE_CHECKING:
    from typing_extensions import TypedDict

    RawAssetTvlBreakdown = tuple[
        spec.Address,
        spec.Address,
        int,
        int,
        int,
        int,
        int,
    ]

    class AssetTvlBreakdown(TypedDict):
        asset: spec.Address
        token: spec.Address
        token_price: int
        underlying_balance: int
        delegated_balance: int
        adjusted_balance: int
        tvl: int

    class ApiVault(TypedDict):
        address: str
        inception: int
        symbol: str
        name: str
        display_name: str
        icon: str
        token: typing.Mapping[str, typing.Any]
        # token.name
        # token.symbol
        # token.address
        # token.decimals
        # token.display_name
        # token.icon
        tvl: typing.Mapping[str, typing.Any]
        # tvl.total_assets
        # tvl.price
        # tvl.tvl
        apy: typing.Mapping[str, typing.Any]
        # apy.type
        # apy.gross_apr
        # apy.net_apy
        # apy.fees
        # apy.fees.performance
        # apy.fees.withdrawal
        # apy.fees.management
        # apy.fees.keep_crv
        # apy.fees.cvx_keep_crv
        # apy.points
        # apy.points.week_ago
        # apy.points.month_ago
        # apy.points.month_ago
        # apy.composite
        # apy.composite.boost
        # apy.composite.pool_apy
        # apy.composite.base_apr
        # apy.composite.boosted_apr
        # apy.composite.cvx_apr
        # apy.composite.rewards_apr
        strategies: typing.Sequence[typing.Mapping[str, str]]
        # strategies[idx].address
        # strategies[idx].name
        endorsed: bool
        version: str
        decimals: int
        type: str
        emergency_shutdown: bool
        updated: int



from __future__ import annotations

import typing

from ctc import spec
from . import erc20_metadata
from . import erc20_state


async def async_print_erc20_summary(
    erc20: spec.Address,
    *,
    include_address: bool = True,
    context: spec.Context = None,
) -> None:
    """print ERC20 summary"""

    import asyncio
    import toolstr
    from ctc import cli

    name_coroutine = erc20_metadata.async_get_erc20_name(
        erc20,
        context=context,
    )
    address_coroutine = erc20_metadata.async_get_erc20_address(
        erc20,
        context=context,
    )
    decimals_coroutine = erc20_metadata.async_get_erc20_decimals(
        erc20, context=context
    )
    symbol_coroutine = erc20_metadata.async_get_erc20_symbol(
        erc20, context=context
    )
    total_supply_coroutine = erc20_state.async_get_erc20_total_supply(
        erc20, context=context
    )

    name, address, decimals, symbol, total_supply = await asyncio.gather(
        name_coroutine,
        address_coroutine,
        decimals_coroutine,
        symbol_coroutine,
        total_supply_coroutine,
    )

    styles = cli.get_cli_styles()

    toolstr.print_text_box('ERC20 Summary for ' + symbol, style=styles['title'])

    rows: typing.MutableSequence[typing.Tuple[str, typing.Any]] = []
    rows.append(('name', name))
    rows.append(('symbol', symbol))
    rows.append(('decimals', decimals))
    rows.append(
        (
            'total_supply',
            toolstr.format(
                total_supply, order_of_magnitude=True, trailing_zeros=True
            ),
        )
    )
    print()
    toolstr.print_table(
        rows,
        # indent=4,
        border=styles['comment'],
        column_styles=[styles['option'], styles['description']],
    )

    if include_address:
        print()
        toolstr.print(address, style=styles['metavar'])




from __future__ import annotations

import os
import typing

if typing.TYPE_CHECKING:
    from typing_extensions import TypedDict

    class Entry(TypedDict):
        id: int
        created_at: str
        text_signature: str
        hex_signature: str
        bytes_signature: str

    class PartialEntry(TypedDict, total=False):
        id: int
        created_at: str
        text_signature: str
        hex_signature: str
        bytes_signature: str


root_url = 'https://www.4byte.directory'

endpoints = {
    'functions': root_url + '/api/v1/signatures/',
    'function_id': root_url + '/api/v1/signatures/{id}',
    'function_hex': root_url
    + '/api/v1/signatures/?hex_signature={hex_signature}',
    'function_text': root_url
    + '/api/v1/signatures/?text_signature={text_signature}',
    #
    'events': root_url + '/api/v1/event-signatures/',
    'event_id': root_url + '/api/v1/event-signatures/{id}',
    'event_hex': root_url
    + '/api/v1/event-signatures/?hex_signature={hex_signature}',
    'event_text': root_url
    + '/api/v1/event-signatures/?text_signature={text_signature}',
}


def get_default_path(datatype: str) -> str:
    import ctc.config

    if datatype not in ['function_signatures', 'event_signatures']:
        raise Exception('unknown datatype: ' + str(datatype))

    return os.path.join(
        ctc.config.get_data_dir(),
        '4byte',
        datatype + '.json',
    )




from __future__ import annotations

import typing
from typing_extensions import TypedDict

from . import context_types
from . import network_types
from . import rpc_types

if typing.TYPE_CHECKING:
    import toolsql
    import toolcli
    import toolstr


class PartialConfig(TypedDict, total=False):
    config_spec_version: str
    data_dir: str
    providers: typing.Mapping[str, rpc_types.Provider]
    networks: typing.Mapping[
        network_types.ChainId, network_types.NetworkMetadata
    ]
    default_network: network_types.ChainId
    default_providers: typing.Mapping[network_types.ChainId, str]
    db_configs: typing.Mapping[str, toolsql.DBConfig]
    context_cache_rules: typing.Sequence[context_types.ContextCacheRule]

    log_rpc_calls: bool
    log_sql_queries: bool

    cli_color_theme: toolcli.StyleTheme
    cli_chart_charset: toolstr.SampleMode


class Config(TypedDict):
    config_spec_version: str
    data_dir: str
    providers: typing.Mapping[str, rpc_types.Provider]
    networks: typing.Mapping[
        network_types.ChainId, network_types.NetworkMetadata
    ]
    default_network: network_types.ChainId
    default_providers: typing.Mapping[network_types.ChainId, str]

    db_configs: typing.Mapping[str, toolsql.DBConfig]
    context_cache_rules: typing.Sequence[context_types.ContextCacheRule]

    log_rpc_calls: bool
    log_sql_queries: bool

    cli_color_theme: toolcli.StyleTheme
    cli_chart_charset: toolstr.SampleMode


class JsonConfig(TypedDict):
    config_spec_version: str
    data_dir: str
    providers: typing.Mapping[str, rpc_types.Provider]
    networks: typing.Mapping[str, network_types.NetworkMetadata]
    default_network: network_types.ChainId
    default_providers: typing.Mapping[str, str]

    db_configs: typing.Mapping[str, toolsql.DBConfig]
    context_cache_rules: typing.Sequence[context_types.ContextCacheRule]

    log_rpc_calls: bool
    log_sql_queries: bool

    cli_color_theme: toolcli.StyleTheme
    cli_chart_charset: toolstr.SampleMode




from __future__ import annotations

import typing

import toolcli
import toolstr

from ctc.protocols import fourbyte_utils


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_fourbyte_command,
        'help': 'lookup 4byte signature',
        'args': [
            {'name': 'signature', 'help': 'signature to look up'},
            {
                'name': '--local',
                'action': 'store_true',
                'help': 'use local database',
            },
            {
                'name': '--remote',
                'action': 'store_true',
                'help': 'use remote database',
            },
        ],
        'examples': [
            '0x18160ddd',
            '0xa9059cbb',
            '"totalSupply()"',
            '"transfer(address,uint256)"',
        ],
    }


async def async_fourbyte_command(
    *, signature: str, local: bool, remote: bool
) -> None:

    if not local and not remote:
        local = True
        remote = True

    try:
        int(signature, 16)
        is_hex = True
    except ValueError:
        is_hex = False

    if is_hex:

        if (signature.startswith('0x') and len(signature) == 10) or len(
            signature
        ) == 8:
            signature_type = 'function'
            results = await fourbyte_utils.async_query_function_signatures(
                hex_signature=signature,
                use_local=local,
                use_remote=remote,
            )

        elif (signature.startswith('0x') and len(signature) == 66) or len(
            signature
        ) == 64:
            signature_type = 'event'
            results = await fourbyte_utils.async_query_event_signatures(
                hex_signature=signature,
                use_local=local,
                use_remote=remote,
            )

        else:
            raise Exception('could not parse signature: ' + str(signature))

    elif '(' in signature and signature.endswith(')'):

        if signature[0].isupper():
            signature_type = 'event'
            results = await fourbyte_utils.async_query_event_signatures(
                text_signature=signature,
                use_local=local,
                use_remote=remote,
            )

        else:
            signature_type = 'function'
            results = await fourbyte_utils.async_query_function_signatures(
                text_signature=signature,
                use_local=local,
                use_remote=remote,
            )
    else:
        raise Exception('could not parse signature: ' + str(signature))

    if signature_type == 'function':
        rows = []
        for result in results:

            hex_signature = result['hex_signature']
            if len(hex_signature) > 10:
                hex_signature = hex_signature[:6] + '...' + hex_signature[-4:]
            row = [
                hex_signature,
                result['text_signature'],
                str(result['id']),
            ]
            rows.append(row)

        labels = [
            'hex',
            'text',
            '4byte id',
        ]

        toolstr.print_text_box('4byte query of ' + str(signature))
        if len(results) == 1:
            print('1 result')
        else:
            print(len(results), 'results')
        print()
        toolstr.print_table(rows, labels=labels)

    elif signature_type == 'event':
        if len(results) == 0:
            print('[no results]')
        elif len(results) > 1:
            raise Exception('event key collision, this should never happen')
        else:
            result = results[0]
            toolstr.print_text_box(result['text_signature'])
            print('- hash:', result['hex_signature'])
            print('- 4byte id:', result['id'])

    else:
        raise Exception('unknown signature_type: ' + str(signature_type))



from __future__ import annotations

import time

import tooltime
import toolstr
import toolcli

from ctc import evm


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_timestamp_command,
        'help': 'get block of timestamp',
        'args': [
            {
                'name': 'timestamp',
                'nargs': '?',
                'help': 'timestamp (default = now)',
            },
        ],
        'examples': ['1642114795'],
    }


async def async_timestamp_command(timestamp: str | int) -> None:
    if timestamp is None:
        timestamp = time.time()

    if not isinstance(timestamp, str):
        raise Exception('function input should be str')

    if len(timestamp) == 4 or (
        len(timestamp) == 10 and timestamp.count('-') == 2
    ):
        pass
    else:
        timestamp = int(timestamp)
    timestamp_seconds = tooltime.timestamp_to_seconds(timestamp)
    if timestamp_seconds > time.time():
        print('[predicting future block]')
        print()
        block = await evm.async_predict_timestamp_block(timestamp=timestamp)
    else:
        block = await evm.async_get_block_of_timestamp(timestamp, mode='<=')
    pretty = tooltime.convert_timestamp(timestamp, 'TimestampISO')

    rows = [
        ['ISO time', pretty],
        ['timestamp', str(timestamp_seconds)],
        ['block before', str(block)],
    ]
    toolstr.print_table(rows, column_justify=['right', 'left'])



from __future__ import annotations

from ctc import spec
from .. import binary_utils
from . import transaction_serialize


def hash_unsigned_transaction(
    transaction: spec.PrechainTransaction,
    *,
    chain_id: int | None = None,
) -> spec.Data:
    """compute hash of unsigned transaction"""

    serialized = transaction_serialize.serialize_unsigned_transaction(
        transaction,
        chain_id=chain_id,
    )
    return binary_utils.keccak(serialized)


def hash_signed_transaction(transaction: spec.PrechainTransaction) -> spec.Data:
    """compute hash of signed transaction"""

    serialized = transaction_serialize.serialize_signed_transaction(transaction)
    return binary_utils.keccak(serialized)



# see https://compound.finance/docs
from __future__ import annotations

import typing

from ctc import rpc
from ctc import spec

blocks_per_day = 6570
days_per_year = 365


async def async_get_supply_apy(
    ctoken: spec.Address,
    *,
    block: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
) -> float:
    result = await rpc.async_eth_call(
        to_address=ctoken,
        function_name='supplyRatePerBlock',
        block_number=block,
        context=context,
    )
    if not isinstance(result, int):
        raise Exception('invalid rpc result')
    supply_rate_per_block = result
    supply_apy = (1 + supply_rate_per_block / 1e18 * blocks_per_day) ** 365 - 1
    return supply_apy


async def async_get_borrow_apy(
    ctoken: spec.Address,
    *,
    block: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
) -> float:
    result = await rpc.async_eth_call(
        to_address=ctoken,
        function_name='borrowRatePerBlock',
        block_number=block,
        context=context,
    )
    if not isinstance(result, int):
        raise Exception('invalid rpc result')
    borrow_rate_per_block = result
    borrow_apy = (1 + borrow_rate_per_block / 1e18 * blocks_per_day) ** 365 - 1
    return borrow_apy


async def async_get_supply_apy_by_block(
    ctoken: spec.Address,
    blocks: typing.Sequence[spec.BlockNumberReference],
    *,
    context: spec.Context = None,
) -> list[float]:

    import numpy as np

    supply_rate_per_block = await rpc.async_batch_eth_call(
        to_address=ctoken,
        function_name='supplyRatePerBlock',
        block_numbers=blocks,
        context=context,
    )
    as_array: spec.NumpyArray = np.array(supply_rate_per_block)
    supply_apy = (1 + as_array / 1e18 * blocks_per_day) ** 365 - 1
    return list(supply_apy)


async def async_get_borrow_apy_by_block(
    ctoken: spec.Address,
    blocks: typing.Sequence[spec.BlockNumberReference],
    *,
    context: spec.Context = None,
) -> list[float]:

    import numpy as np

    borrow_rate_per_block = await rpc.async_batch_eth_call(
        to_address=ctoken,
        function_name='borrowRatePerBlock',
        block_numbers=blocks,
        context=context,
    )
    as_array: spec.NumpyArray = np.array(borrow_rate_per_block)
    borrow_apy = (1 + as_array / 1e18 * blocks_per_day) ** 365 - 1
    return list(borrow_apy)



from __future__ import annotations

from .. import query_utils
from . import dex_pools_statements


async_query_dex_pool = query_utils.wrap_selector_with_connection(
    dex_pools_statements.async_select_dex_pool,
    'dex_pools',
)


async_query_dex_pools = query_utils.wrap_selector_with_connection(
    dex_pools_statements.async_select_dex_pools,
    'dex_pools',
)

async_query_dex_pool_factory_last_scanned_block = (
    query_utils.wrap_selector_with_connection(
        dex_pools_statements.async_select_dex_pool_factory_last_scanned_block,
        'dex_pools',
    )
)



from __future__ import annotations

import typing

import toolcli

from ctc import evm
from ctc import spec
from ctc.cli import cli_utils


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_events_command,
        'help': 'get contract events',
        'args': [
            {
                'name': 'contract',
                'help': 'contract address of event',
            },
            {
                'name': 'event',
                'help': 'event name or event hash',
            },
            {
                'name': '--blocks',
                'help': 'block range',
            },
            {
                'name': '--include-timestamps',
                'default': False,
                'action': 'store_true',
                'help': 'include timestamps',
            },
            {
                'name': '--export',
                'default': 'stdout',
                'help': 'file path for output (.json or .csv)',
            },
            {
                'name': '--overwrite',
                'action': 'store_true',
                'help': 'specify that output path can be overwritten',
            },
            {
                'name': ['--verbose', '-v'],
                'help': 'display more event data',
                'action': 'store_true',
            },
        ],
        'examples': {
            '0x956f47f50a910163d8bf957cf5846d573e7f87ca Transfer': {
                'long': True
            },
            '0x956f47f50a910163d8bf957cf5846d573e7f87ca Transfer --blocks 14000000:14010000': {},
            '0x956f47f50a910163d8bf957cf5846d573e7f87ca Transfer --blocks 14000000:14010000 --include-timestamps': {},
        },
    }


async def async_events_command(
    *,
    contract: str,
    event: str,
    blocks: str,
    include_timestamps: bool,
    export: str,
    overwrite: bool,
    verbose: bool,
) -> None:

    import polars as pl

    contract = await evm.async_resolve_address(contract)

    if blocks is not None:
        all_blocks = await cli_utils.async_parse_block_range(blocks)
        start_block = all_blocks[0]
        end_block = all_blocks[-1]
    else:
        start_block = None
        end_block = None

    if event.startswith('0x'):
        events: spec.DataFrame = await evm.async_get_events(
            contract_address=contract,
            start_block=start_block,
            end_block=end_block,
            include_timestamps=include_timestamps,
            verbose=False,
            event_hash=event,
        )
    else:
        events = await evm.async_get_events(
            contract_address=contract,
            start_block=start_block,
            end_block=end_block,
            include_timestamps=include_timestamps,
            verbose=False,
            event_name=event,
        )

    if len(events) == 0:
        print('[no events found]')

    else:

        # output
        if not verbose:
            columns = [
                column
                for column in events.columns
                if column.startswith('arg__')
            ]
            if include_timestamps:
                columns.insert(0, 'timestamp')

            events = events[columns]
            new_column_names = {
                old_column: old_column[5:]
                for old_column in events.columns
                if old_column.startswith('arg__')
            }
            events = events.rename(new_column_names)
        if export == 'stdout' and include_timestamps:
            events = events.with_columns(pl.col('timestamp').cast(pl.Utf8))
        cli_utils.output_data(events, output=export, overwrite=overwrite)



from __future__ import annotations

import typing


if typing.TYPE_CHECKING:
    from typing_extensions import TypedDict, NotRequired

    class Trade(TypedDict):
        x_bought: int | float
        x_sold: int | float
        y_bought: int | float
        y_sold: int | float
        fee_rate: float
        new_pool: NewPoolState

    class Mint(TypedDict):
        x_deposited: int | float
        y_deposited: int | float
        lp_minted: int | float
        new_pool: NewPoolState

    class Burn(TypedDict):
        x_withdrawn: int | float
        y_withdrawn: int | float
        lp_burned: int | float
        new_pool: NewPoolState

    class NewPoolState(TypedDict):
        x_reserves: int | float
        y_reserves: int | float
        lp_total_supply: NotRequired[int | float]

    class TradeSummary(TypedDict):
        end_slippage_x_per_y: int | float
        end_slippage_y_per_x: int | float
        mean_slippage_x_per_y: int | float
        mean_slippage_y_per_x: int | float
        mean_x_per_y: int | float
        mean_y_per_x: int | float
        x_per_y_start: int | float
        y_per_x_start: int | float
        x_per_y_end: int | float
        y_per_x_end: int | float
        x_fees: int | float
        y_fees: int | float
        trade_results: Trade




from __future__ import annotations

from ctc import db

from . import chainlink_statements


async_query_feed = db.query_utils.wrap_selector_with_connection(
    chainlink_statements.async_select_feed,
    'chainlink',
)

async_query_feeds = db.query_utils.wrap_selector_with_connection(
    chainlink_statements.async_select_feeds,
    'chainlink',
)

async_query_aggregator_updates = db.query_utils.wrap_selector_with_connection(
    chainlink_statements.async_select_aggregator_updates,
    'chainlink',
)



from __future__ import annotations

import copy
import typing

import toolsql

from ctc import config
from ctc import spec
from ctc.spec.typedefs import db_types

from . import schemas

#
# # schema lists
#


def get_admin_schema_names() -> tuple[db_types.AdminSchemaName]:
    return db_types.AdminSchemaName.__args__  # type: ignore


def get_generic_schema_names() -> tuple[db_types.GenericSchemaName]:
    return db_types.GenericSchemaName.__args__  # type: ignore


def get_network_schema_names() -> tuple[db_types.NetworkSchemaName]:
    return db_types.NetworkSchemaName.__args__  # type: ignore


def get_all_schema_names() -> typing.Sequence[db_types.SchemaName]:
    schema_names = (
        get_admin_schema_names()
        + get_generic_schema_names()
        + get_network_schema_names()
    )
    return typing.cast(typing.Sequence[db_types.SchemaName], schema_names)


#
# # raw schemas
#


def get_raw_schema(schema_name: str) -> toolsql.DBSchema:
    """get raw schema, a set of tables not customized for any network"""
    schema: toolsql.DBSchemaShorthand
    if schema_name == 'block_timestamps':
        schema = schemas.block_timestamps_schema
    elif schema_name == 'block_gas':
        schema = schemas.block_gas_schema
    elif schema_name == 'blocks':
        schema = schemas.blocks_schema
    elif schema_name == 'contract_abis':
        schema = schemas.contract_abis_schema
    elif schema_name == 'contract_creation_blocks':
        schema = schemas.contract_creation_blocks_schema
    elif schema_name == 'dex_pools':
        schema = schemas.dex_pools_schema
    elif schema_name == 'erc20_metadata':
        schema = schemas.erc20_metadata_schema
    elif schema_name == 'events':
        schema = schemas.events_schema
    # elif schema_name == 'erc20_state':
    #     schema = schemas.erc20_state_schema
    elif schema_name == 'schema_versions':
        schema = schemas.schema_versions_schema
    elif schema_name == 'transactions':
        schema = schemas.transactions_schema
    #
    # # protocols
    #
    elif schema_name == '4byte':
        from ctc.protocols import fourbyte_utils

        schema = fourbyte_utils.fourbyte_schema
    elif schema_name == 'chainlink':
        from ctc.protocols.chainlink_utils import chainlink_db

        schema = chainlink_db.chainlink_schema
    elif schema_name == 'coingecko':
        from ctc.protocols.coingecko_utils import coingecko_db

        schema = coingecko_db.coingecko_schema
    else:
        raise Exception('unknown schema: ' + str(schema_name))

    return toolsql.normalize_shorthand_db_schema(schema)


#
# # prepared schemas
#


def get_prepared_schema(
    schema_name: str,
    context: spec.Context = None,
) -> toolsql.DBSchema:
    """get prepared schema, a set of tables customized for a specific network"""

    # get schema
    schema = get_raw_schema(schema_name)
    schema = copy.deepcopy(schema)

    # add network to table name
    for table_name, table in list(schema['tables'].items()):
        full_name = get_prepared_table_name(
            context=context, table_name=table_name
        )
        if table.get('name') is not None:
            table['name'] = full_name
        schema['tables'][full_name] = schema['tables'].pop(table_name)  # type: ignore

    return schema


def get_table_schema(
    table_name: str,
    context: spec.Context = None,
) -> toolsql.TableSchema:
    schema_name = _get_schema_of_raw_table(table_name)
    if schema_name in get_generic_schema_names():
        schema = get_raw_schema(schema_name=schema_name)
        return schema['tables'][table_name]
    else:
        schema = get_prepared_schema(schema_name=schema_name, context=context)
        prepared_name = get_prepared_table_name(
            context=context, table_name=table_name
        )
        return schema['tables'][prepared_name]


def get_prepared_table_name(
    table_name: str,
    context: spec.Context = None,
) -> str:
    """get full table name, incorporating chain information"""
    chain_id = config.get_context_chain_id(context)
    return 'network_' + str(chain_id) + '__' + table_name


def _get_schema_of_raw_table(table: str) -> spec.SchemaName:
    candidates = []
    for schema_name in get_all_schema_names():
        raw_schema = get_raw_schema(schema_name)
        for schema_table in raw_schema['tables'].keys():
            if table == schema_table:
                candidates.append(schema_name)

    if len(candidates) == 1:
        return candidates[0]
    elif len(candidates) == 0:
        raise Exception(
            'could not find any schema containing table ' + str(table)
        )
    else:
        raise Exception('found multiple schemas containing table ' + str(table))




from __future__ import annotations

import typing

from ctc import evm
from ctc import rpc
from ctc import spec

from . import ens_directory
from . import registrar


def hash_name(name: str) -> spec.PrefixHexData:
    import idna

    labels = name.split('.')
    output = '00' * 32
    for label in labels[::-1]:
        label_bytes = idna.encode(label, uts46=True)
        label_hash = evm.keccak_text(label_bytes, output_format='raw_hex')
        output = evm.keccak(output + label_hash, output_format='raw_hex')
    return '0x' + output


async def async_resolve_name(
    name: str,
    *,
    block: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
) -> spec.Address | None:
    name_hash = hash_name(name)

    function_abi: spec.FunctionABI = {
        'name': 'addr',
        'inputs': [{'type': 'bytes32'}],
        'outputs': [{'type': 'address'}],
    }

    result = await rpc.async_eth_call(
        to_address=ens_directory.resolver,
        block_number=block,
        function_abi=function_abi,
        function_parameters=[name_hash],
        context=context,
    )
    if not isinstance(result, str):
        raise Exception('invalid rpc result')

    if result == '0x0000000000000000000000000000000000000000':
        resolver = await registrar.async_get_resolver(
            name=name, block=block, context=context
        )
        if resolver == '0x0000000000000000000000000000000000000000':
            return None
        result = await rpc.async_eth_call(
            to_address=resolver,
            block_number=block,
            function_abi={
                'name': 'addr',
                'inputs': [{'type': 'bytes32'}],
                'outputs': [{'type': 'address'}],
            },
            function_parameters=[name_hash],
            context=context,
        )
        if not isinstance(result, str):
            raise Exception('invalid rpc result')

    return result


async def async_resolve_names(
    names: typing.Sequence[str],
    *,
    block: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
) -> typing.Sequence[spec.Address | None]:

    import asyncio

    coroutines = [
        async_resolve_name(name=name, block=block, context=context)
        for name in names
    ]

    return await asyncio.gather(*coroutines)


async def async_reverse_lookup(
    address: spec.Address,
    *,
    block: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
) -> str:
    function_abi: spec.FunctionABI = {
        'name': 'getNames',
        'inputs': [{'type': 'address[]'}],
        'outputs': [{'type': 'string[]'}],
    }
    names = await rpc.async_eth_call(
        to_address=ens_directory.reverse_records,
        block_number=block,
        function_abi=function_abi,
        function_parameters=[[address]],
        context=context,
    )
    output = names[0]
    if not isinstance(output, str):
        raise Exception('invalid rpc result')
    return output


async def async_name_history(
    *,
    context: spec.Context = None,
) -> None:
    raise NotImplementedError()


async def async_get_text_record(
    key: str,
    *,
    name: str | None = None,
    node: str | None = None,
    context: spec.Context = None,
) -> str:

    if node is None:
        if name is None:
            raise Exception('must specify name or node')
        node = hash_name(name)

    function_abi: spec.FunctionABI = {
        'name': 'text',
        'inputs': [
            {'name': 'node', 'type': 'bytes32'},
            {'name': 'key', 'type': 'string'},
        ],
        'outputs': [{'name': '', 'type': 'string'}],
    }

    result = await rpc.async_eth_call(
        to_address=ens_directory.resolver,
        function_abi=function_abi,
        function_parameters=[node, key],
        context=context,
    )
    if not isinstance(result, str):
        raise Exception('invalid rpc result')
    return result


async def async_get_text_records(
    *,
    name: str | None = None,
    node: str | None = None,
    keys: typing.Sequence[str] | None = None,
    context: spec.Context = None,
) -> dict[str, str]:
    """
    https://docs.ens.domains/ens-improvement-proposals/ensip-5-text-records
    """
    import asyncio

    if node is None:
        if name is None:
            raise Exception('must specify name or node')
        node = hash_name(name)

    if keys is None:
        text_changes = await async_get_text_changes(
            name=name, node=node, context=context
        )
        keys = text_changes['arg__key'].to_list()

    coroutines = [
        async_get_text_record(key=key, node=node, context=context)
        for key in keys
    ]
    values = await asyncio.gather(*coroutines)
    return dict(zip(keys, values))


async def async_get_text_changes(
    *,
    name: str | None = None,
    node: str | None = None,
    context: spec.Context = None,
) -> spec.DataFrame:

    event_abi: spec.EventABI = {
        'name': 'TextChanged',
        'type': 'event',
        'inputs': [
            {
                'indexed': True,
                'internalType': 'bytes32',
                'name': 'node',
                'type': 'bytes32',
            },
            {
                'indexed': True,
                'internalType': 'string',
                'name': 'indexedKey',
                'type': 'string',
            },
            {
                'indexed': False,
                'internalType': 'string',
                'name': 'key',
                'type': 'string',
            },
        ],
    }

    events = await evm.async_get_events(
        contract_address=ens_directory.resolver,
        event_abi=event_abi,
        start_block=9000000,
        verbose=2,
        context=context,
        convert_invalid_str_to_none=True,
    )

    if node is None:
        if name is None:
            raise Exception('must specify name or node')
        node = hash_name(name)

    mask = events['arg__node'] == node
    return events.filter(mask)


async def async_get_content_hash(
    name: str | None = None,
    node: str | None = None,
    *,
    context: spec.Context = None,
) -> str:

    if node is None:
        if name is None:
            raise Exception('must specify name or node')
        node = hash_name(name)

    function_abi: spec.FunctionABI = {
        'name': 'contenthash',
        'inputs': [
            {'name': 'node', 'type': 'bytes32'},
        ],
        'outputs': [{'name': '', 'type': 'bytes'}],
    }

    result = await rpc.async_eth_call(
        to_address=ens_directory.resolver,
        function_abi=function_abi,
        function_parameters=[node],
        context=context,
    )
    if not isinstance(result, str):
        raise Exception('invalid rpc result')
    return result


async def async_get_expiration(
    name: str,
    *,
    context: spec.Context = None,
) -> int:

    if not name.endswith('.eth'):
        raise NotImplementedError('only implemented for .eth domains')

    label = name.split('.')[-2]
    label_id = evm.keccak_text(label, output_format='integer')

    function_abi: spec.FunctionABI = {
        'name': 'nameExpires',
        'inputs': [{'name': 'id', 'type': 'uint256'}],
        'outputs': [{'type': 'uint256'}],
    }

    result = await rpc.async_eth_call(
        to_address=ens_directory.base_registrar,
        function_abi=function_abi,
        function_parameters=[label_id],
        context=context,
    )
    if not isinstance(result, int):
        raise Exception('invalid rpc result')
    return result




from __future__ import annotations

from .. import query_utils
from . import erc20_metadata_statements


async_query_erc20_metadata = query_utils.wrap_selector_with_connection(
    erc20_metadata_statements.async_select_erc20_metadata,
    'erc20_metadata',
)


async_query_erc20s_metadata = query_utils.wrap_selector_with_connection(
    erc20_metadata_statements.async_select_erc20s_metadata,
    'erc20_metadata',
)



from __future__ import annotations

from ctc import db
from . import coingecko_statements


async_query_token = db.wrap_selector_with_connection(
    coingecko_statements.async_select_token,
    'coingecko',
)

async_query_tokens = db.wrap_selector_with_connection(
    coingecko_statements.async_select_tokens,
    'coingecko',
)



from __future__ import annotations

import toolcli
import toolstr

from ctc import cli
from ctc.cli import cli_utils


def get_command_spec() -> toolcli.CommandSpec:
    styles = cli.get_cli_styles()
    return {
        'f': color_command,
        'help': 'display or set color themes',
        'args': [
            {
                'name': 'theme',
                'help': 'color theme to set, '
                + toolstr.add_style('default', styles['option'])
                + ' to reset, '
                + toolstr.add_style('custom', styles['option'])
                + ' to customize',
                'nargs': '?',
            },
            {
                'name': ['-v', '--verbose'],
                'help': 'display available color themes',
                'action': 'store_true',
            },
        ],
        'examples': [
            '',
            'nord',
            '-v',
        ],
    }


def color_command(theme: str | None, verbose: bool) -> None:

    styles = cli.get_cli_styles()

    if theme is not None:
        if theme in ['default', 'reset']:
            cli_utils.reset_color_theme()
        elif theme == 'custom':
            print('setting custom color theme')
            styles = cli_utils.create_custom_color_theme()
            cli_utils.set_color_theme(styles)
        else:
            print('setting color theme to:', theme)
            styles = cli_utils.set_color_theme(theme)
        toolstr.print_text_box('Color Theme', style=styles['title'])
        cli_utils.print_color_theme(styles)

    elif verbose:
        cli_utils.preview_color_themes()
        print()
        toolstr.print(
            'enable a theme with '
            + toolstr.add_style(
                'ctc color <theme_name>', styles['option'] + ' bold'
            ),
            style=styles['comment'],
        )

    else:
        toolstr.print_text_box('Color Theme', style=styles['title'])
        cli_utils.print_color_theme(styles)
        print()
        toolstr.print(
            'to view available color themes, use '
            + toolstr.add_style('ctc color -v', styles['option']),
            style=styles['comment'],
        )
        print()
        toolstr.print(
            'to set a color theme, use '
            + toolstr.add_style('ctc color <theme>', styles['option']),
            style=styles['comment'],
        )
        toolstr.print(
            '    where '
            + toolstr.add_style('<theme>', styles['option'])
            + ' is a theme name or theme specification',
            style=styles['comment'],
        )
        print()
        toolstr.print(
            'to reset to the default color theme, use '
            + toolstr.add_style('ctc color default', styles['option']),
            style=styles['comment'],
        )



from __future__ import annotations

import asyncio
import typing

import toolsql

from ctc import spec
from ... import management
from ... import schema_utils
from . import block_timestamps_statements

from .multischema_block_timestamps_statements import (
    async_select_block_timestamp,
)

if typing.TYPE_CHECKING:
    from typing_extensions import Literal


async def async_select_timestamp_block(
    timestamp: int,
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
    mode: Literal['<=', '>=', '=='] = '>=',
) -> int | None:
    """

    possible approaches to ensure db has enough data to give answer:
    - possibilities
        - run two queries, like (gte + lt) or (lte + gt)
            - limit = 1 on both
            - make sure that the results are only one apart
        - run one query first, then...
            - run second query to make sure result+1 or result-1 is in db
        - assume db is complete, do not perform additional checks
    - the relative speeds of these operations matter
        - this all could be bikeshedding
    """

    if mode == '<=':
        query = {
            'where_lte': {'timestamp': timestamp},
            'order_by': {'column': 'block_number', 'order': 'descending'},
        }
    elif mode == '>=':
        query = {'where_gte': {'timestamp': timestamp}}
    elif mode == '==':
        query = {'where_equals': {'timestamp': timestamp}}
    else:
        raise Exception('unknown mode: ' + str(mode))

    timestamp_schema = management.get_active_timestamp_schema()

    if timestamp_schema == 'block_timestamps':
        table = schema_utils.get_table_schema(
            'block_timestamps', context=context
        )
        block_numbers: typing.Sequence[int] = await toolsql.async_select(
            conn=conn,
            table=table,
            columns=['block_number'],
            output_format='single_column',
            **query,
        )
        if block_numbers is not None and len(block_numbers) > 0:
            block_number = block_numbers[0]
        else:
            block_number = None
    elif timestamp_schema == 'blocks':
        raise NotImplementedError()
    else:
        raise Exception('unknown schema: ' + str(timestamp_schema))

    if block_number is not None:
        if mode == '==':
            pass
        elif mode == '<=':
            # assert block after is in db
            next_timestamp = await async_select_block_timestamp(
                conn=conn,
                block_number=block_number + 1,
            )
            if next_timestamp is None:
                return None
        elif mode == '>=':
            # assert block before is in db
            previous_timestamp = await async_select_block_timestamp(
                conn=conn,
                block_number=block_number - 1,
            )
            if previous_timestamp is None:
                return None
        else:
            raise Exception('unknown mode: ' + str(mode))

    return block_number


async def async_select_timestamps_blocks(
    timestamps: typing.Sequence[int],
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
    mode: Literal['<=', '>=', '=='] = '>=',
) -> list[int | None]:

    timestamp_schema = management.get_active_timestamp_schema()

    if timestamp_schema == 'block_timestamps':
        coroutines = [
            async_select_timestamp_block(
                conn=conn,
                timestamp=timestamp,
                context=context,
                mode=mode,
            )
            for timestamp in timestamps
        ]
        return await asyncio.gather(*coroutines)

    elif timestamp_schema == 'blocks':
        raise NotImplementedError()

    else:
        raise Exception('unknown schema: ' + str(timestamp_schema))


async def async_select_timestamp_block_range(
    timestamp: int,
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> tuple[int | None, int | None]:

    timestamp_schema = management.get_active_timestamp_schema()

    if timestamp_schema == 'block_timestamps':
        return await block_timestamps_statements.async_select_timestamp_block_range(
            timestamp=timestamp,
            conn=conn,
            context=context,
        )
    elif timestamp_schema == 'blocks':
        raise NotImplementedError()
    else:
        raise Exception('unknown schema: ' + str(timestamp_schema))

